/* 
 * module: global.armci.c
 * author: Jarek Nieplocha
 * description: implements GA primitive operations --
 *              create (regular& irregular) and duplicate, destroy, get, put,
 *              accumulate, scatter, gather, read&increment & synchronization 
 * 
 * DISCLAIMER
 *
 * This material was prepared as an account of work sponsored by an
 * agency of the United States Government.  Neither the United States
 * Government nor the United States Department of Energy, nor Battelle,
 * nor any of their employees, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
 * ASSUMES ANY LEGAL LIABILITY OR RESPONSIBILITY FOR THE ACCURACY,
 * COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
 * SOFTWARE, OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT
 * INFRINGE PRIVATELY OWNED RIGHTS.
 *
 *
 * ACKNOWLEDGMENT
 *
 * This software and its documentation were produced with United States
 * Government support under Contract Number DE-AC06-76RLO-1830 awarded by
 * the United States Department of Energy.  The United States Government
 * retains a paid-up non-exclusive, irrevocable worldwide license to
 * reproduce, prepare derivative works, perform publicly and display
 * publicly by or for the US Government, including the right to
 * distribute to other US Government contractors.
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

extern "C" {
  #include "global.h"
  #include "globalp.h"
  #include "message.h"
  #include "macdecls.h"
}

#include "MPIinclude.h"
#include "globals.h"
#include "Procs.h"
#include "Range.h"
#include "DAD.h"

#include "global.armci.h"

#define DEBUG 0
#define USE_MALLOC 1
#define INVALID_MA_HANDLE -1 
#define NEAR_INT(x) (x)< 0.0 ? ceil( (x) - 0.5) : floor((x) + 0.5)

/*uncomment line below to initialize arrays in ga_create/duplicate */
/*#define GA_CREATE_INDEF yes */

/*uncomment line below to verify consistency of MA in every sync */
/*#define CHECK_MA yes */

/* uncomment line below to verify if MA base address is alligned wrt datatype*/ 
#if !(defined(LINUX) || defined(CRAY))
#define CHECK_MA_ALGN 1
#endif


// Not sure what this is for (dbc)
//#define FATR


// *** customized Adlib code to support GA ***

// `processArrays' table is used to ensure that if several process
// grids are defined with the same shape they refer to the same
// `Procs' structure.

// Construction and destruction.

struct Ad_ProcessArrayRecord {
  int rank ;
  int shape [7] ;
  Procs* procs ;
} ;

#define MAX_PROCESS_ARRAYS 100

static Ad_ProcessArrayRecord processArrays [MAX_PROCESS_ARRAYS] ;
static int numProcessArrays = 0 ;

static Procs* getProcs(int rank, int shape []) {

  // Scan `processArrays' table for pre-existing process array of the
  // same shape.  Return that if it exists.

  for(int slot = 0 ; slot < numProcessArrays ; slot++) {
    Ad_ProcessArrayRecord& processArray = processArrays [slot] ;
    if(rank == processArray.rank) {
      int i = 0 ;
      while (i < rank && shape [i] == processArray.shape [i]) i++ ;
      if(i == rank) return processArray.procs;
    }
  }

  // Otherwise create a new `Procs'

  if(numProcessArrays == MAX_PROCESS_ARRAYS) {
    fprintf(stderr, "** RUNTIME ERROR **  Process array records exhausted\n") ;
    exit(1) ;
    return 0 ;
  }
  else {
    int i ;

    // Check size of processor array is number of available processors.

    int size = 1 ;
    for (i = 0 ; i < rank ; i++) {
      size *= shape [i] ;
    }

    int asize = apg.size() ;

    if(size != asize && size != 1) {
      fprintf(stderr, "** RUNTIME ERROR **  "
              "Processor array requires %d processes---only %d available\n",
              size, asize) ;
      exit(1) ;
    }

    // Create new process array, and add it to `processArrays' table.

    Ad_ProcessArrayRecord& processArray = processArrays [numProcessArrays] ;

    processArray.rank = rank ;
    for (i = 0 ; i < rank ; i++)
      processArray.shape [i] = shape [i] ;
    processArray.procs = new Procs(rank, shape) ;

    numProcessArrays++ ;

    return processArray.procs ;
  }
}


// Patch access schedules

#define MBLKT_REM 0    // Remote processor id.
#define MBLKT_SIZ 1    // Number of bytes in block
#define MBLKT_NST 2    // Number of stride levels.
#define MBLKT_CTG 3    // Length of contiguous blocks in bytes
#define MBLKT_SHP 4    // First of `NST' extents (followed by `NST' strides
                       // for 1st block, then `NST' strides for 2nd block).

// NB: Unit of strides is one byte (*not* one array element).
// NB: If these constants are changed, `transReq' below must be changed.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    `Vec': A lightweight ``Vector'' class.                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T, int iniCapacity>
class Vec {
public :
  Vec() : n(0), capacity(iniCapacity), vals(iniVals) {}
  ~Vec() {if(vals != iniVals) delete [] vals ; }

  void add(T val) {
    if(n == capacity) extend(2 * capacity) ;
    vals [n] = val ;
    n++ ;
  }

  T* add() {
    if(n == capacity) extend(2 * capacity) ;
    T* result = vals + n ;
    n++ ;
    return result ;
  }

  T* addSome(const int d) {
    const int newN = n + d ;
    if(newN > capacity) {
      const int newCapacity = 2 * capacity ;
      extend(newN > newCapacity ? newN : newCapacity) ;
    }
    T* ptr = vals + n ;
    n = newN ;
    return ptr ;
  }

  void clear() {n = 0 ; }

  T* last() {return vals + n - 1 ; }

  int size() {return n ; }
  T* data() {return vals ; }
private :
  Vec(const Vec<T, iniCapacity>& v) {} ;  // forbid copying

  T iniVals [iniCapacity] ;

  int n, capacity ;
  T* vals ;

  void extend(const int newCapacity) {

    // newCapacity >= n

    T* newVals = new T [newCapacity] ;
    memcpy((char*) newVals, (char*) vals, sizeof(T) * n) ;
    if(vals != iniVals) delete [] vals ;
    vals     = newVals ;
    capacity = newCapacity ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockRMSchedule' class.                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockRMSchedule {
public :
  BlockRMSchedule(void* _locBase, const int _remOff, void** _remBases,
                  const int _rank, const int _elementLen) :
               locBase(_locBase), remOff(_remOff), remBases(_remBases),
               rank(_rank), elementLen(_elementLen),
               blockLen(MBLKT_SHP + 3 * _rank) {
    numBlk = 0 ;
  }

  void transReq(int locOffset, int* locStrs,
		int remOffset, int* remStrs,
                int* exts, int numStrLevels, int contigLen,
                int vol, int remId) {

    // Add a transfer request to the appropriate lists.

    locPtrs.add((char*) locBase + elementLen * locOffset) ;
    remPtrs.add((char*) remBases [remId] + elementLen * (remOff + remOffset)) ;

    int* block = blocks.addSome(blockLen) ;

    *(block++) = remId ;
    *(block++) = elementLen * vol ;
    *(block++) = numStrLevels ;
    *(block++) = elementLen * contigLen ;

    int* bexts  = block ;
    int* bstrs1 = bexts + numStrLevels ;
    int* bstrs2 = bstrs1 + numStrLevels ;
    for(int i = 0 ; i < numStrLevels ; i++) {
      *(bexts++)  = *(exts++) ;
      *(bstrs1++) = elementLen * *(locStrs++) ;
      *(bstrs2++) = elementLen * *(remStrs++) ;
    }

    numBlk++ ;
  }

  void** getLocs() {return locPtrs.data() ; }
  void** getRems() {return remPtrs.data() ; }
  int*   getBlks() {return blocks.data()  ; }

  const int rank ;

  const int blockLen ;
  const int elementLen ;
  int numBlk ;

private :

  void* locBase ;
  int remOff ;
  void** remBases ;

  Vec<void*, 16> locPtrs, remPtrs ;
  Vec<int*,  16> blkPtrs ;

  Vec<int, 16> blocks ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PatchTransfers': loop state for enumeration in `nga_put_',   //
//                         etc.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Optimized from original version which used standard `AllBlocksIndex'.
// This version doesn't support block-cyclic...
// (Also cut use of `Group' to define target of remote operation.)

class PatchTransfers : public BlockRMSchedule {
public :
  PatchTransfers(const Integer _ld [], void* locDat,
                 const DAD* _rem, void** remPtrs, const int elementLen,
                 Integer _lo [], Integer _hi []) :
              BlockRMSchedule(locDat, 0, remPtrs, _rem->rnk(), elementLen),
              ld(_ld), rem(_rem), rank(_rem->rnk()),
              lo(_lo), hi(_hi), remPrc(_rem->grp().prc()) {}

  void loop(const int locOffset, const int locCtg,
            const int locStep,
            const int remOffset, const int remCtg,
            const int vol,
            const int remId,
            const int r) ;

private :
  int exts [MAXDIM], locSteps [MAXDIM], remSteps [MAXDIM] ;

  const Integer* ld ;

  const DAD* rem ;
  const Procs* remPrc ;

  int rank ;

  Integer *lo, *hi ;
} ;


void PatchTransfers :: loop(const int locOffset, const int locStep,
                            const int remOffset,
                            const int numStrLevels, const int contigLen,
                            const int vol,
                            const int remId,
                            const int r) {

  if(r == rank)
    transReq(locOffset, locSteps, remOffset, remSteps, exts,
             numStrLevels, contigLen,
             vol, remPrc->id_abs(remId)) ;
  else {

    Map remMap = rem->map(r) ;
    const int len = remMap.str() ;
    Range y = remMap.rng() ;
    Dimension d = y.dim() ;

    int x_bas = lo [r] - 1, x_size = hi [r] - lo [r] + 1 ;

    int d_size, d_bas, d_str ;
    y.crds(&d_size, &d_bas, &d_str, x_size, x_bas) ;

    Block b ;

    int crd = d_bas ;

    for(int d_l = 0 ; d_l < d_size ; d_l++) {

      y.block(&b, crd, x_size, x_bas) ;

      //const int remOffsetElem = remOffset + remMap.disp(b.sub_bas) ;
      //const int remStep       = remMap.step(b.sub_stp) ;
      const int remOffsetElem = remOffset + len * b.sub_bas ;
      const int remStep       = len * b.sub_stp ;
            // Lose support for packed or extended ranges.

      const int jlocStep = locStep * b.glb_stp ;

      if(numStrLevels == 0 && jlocStep == contigLen && remStep == contigLen)
	loop(locOffset + b.glb_bas * locStep, locStep * ld [r],
	     remOffsetElem,
	     0, contigLen * b.count,
	     vol * b.count,
	     remId + crd * d.str_prc(),
	     r + 1) ;
      else {
	exts     [numStrLevels] = b.count ;
	locSteps [numStrLevels] = jlocStep ;
	remSteps [numStrLevels] = remStep ;

	loop(locOffset + b.glb_bas * locStep, locStep * ld [r],
	     remOffsetElem,
	     numStrLevels + 1, contigLen,
	     vol * b.count,
	     remId + crd * d.str_prc(),
	     r + 1) ;
      }

      crd += d_str ;
    }
  }
}



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PatchOwners': loop state for enumeration in                  //
//                      `nga_locate_region_'                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class PatchOwners {
public :
  PatchOwners(const DAD* _rem, Integer* __lo, Integer* __hi,
              Integer* _map, Integer* _proclist) :
      rem(_rem), lo(__lo), hi(__hi), rank(_rem->rnk()),
      map(_map), proclist(_proclist) {}

  void loop(Group remGrp, const int r, Integer* np) ;

private :
  Integer *hi, *lo ;

  const DAD* rem ;
  Integer *map, *proclist ;

  int rank ;

  int _hi [MAXDIM], _lo [MAXDIM] ;
} ;

void PatchOwners :: loop(Group remGrp, const int r, Integer* np) {

  if(r == rank) {

    for(int d = 0; d < rank; d++) {
      map [d]        = _lo [d] ;
      map [rank + d] = _hi [d] ;
    }

    *proclist = remGrp.lead_abs() ;

    map += rank * 2 ;
    proclist++ ;

    (*np)++;
  }
  else {
    Range y = rem->rng(r) ;

    AllBlocksIndex j(y.triplet(lo [r] - 1, hi [r] - 1)) ;

    for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {
      _lo [r] = j.glb_bas + 1 ;
      _hi [r] = j.glb_bas + j.count ;

      loop(remGrp / j, r + 1, np) ;
    }
  }
}

// *** (end of customized Adlib code to support GA) ***



/*\ SYNCHRONIZE ALL THE PROCESSES
\*/
void FATR ga_sync_()
{
extern int GA_fence_set;
#ifdef CHECK_MA
Integer status;
#endif

       GA_fence_set=0;
       ARMCI_AllFence();
       ga_msg_sync_();
#ifdef CHECK_MA
       status = MA_verify_allocator_stuff();
#endif
}


Integer GAsizeof(Integer type)    
{
  switch (type) {
     case MT_F_DBL  : return (sizeof(DoublePrecision));
     case MT_F_INT  : return (sizeof(Integer));
     case MT_F_DCPL : return (sizeof(DoubleComplex));
          default   : return 0; 
  }
}


/*\ FINAL CLEANUP of shmem when terminating
\*/
void ga_clean_resources()
{
    ARMCI_Cleanup();
}


/*\ CHECK GA HANDLE and if it's wrong TERMINATE
 *  Fortran version
\*/
#if defined(CRAY) || defined(WIN32)
void FATR  ga_check_handle_(Integer* g_a, _fcd fstring)
#else
void FATR  ga_check_handle_(Integer* g_a, char* fstring, int slen)
#endif
{
char  buf[FLEN];

    if( GA_OFFSET + (*g_a) < 0 || GA_OFFSET + (*g_a) >= max_global_array){
#if defined(CRAY) || defined(WIN32)
      f2cstring(_fcdtocp(fstring), _fcdlen(fstring), buf, FLEN);
#else
      f2cstring(fstring, slen, buf, FLEN);
#endif
      fprintf(stderr, " ga_check_handle: %s ", buf);
      ga_error(" invalid global array handle ", (*g_a));
    }
    if( ! (GA[GA_OFFSET + (*g_a)].actv) ){
#if defined(CRAY) || defined(WIN32)
      f2cstring(_fcdtocp(fstring), _fcdlen(fstring), buf, FLEN);
#else
      f2cstring(fstring, slen, buf, FLEN);
#endif
      ga_error(" global array is not active ", (*g_a));
    }
}



/*\ CHECK GA HANDLE and if it's wrong TERMINATE
 *  C version
\*/
void ga_check_handle(Integer *g_a,char * string)
{
  ga_check_handleM(g_a,string);
}




/*\ prepare permuted list of processes for remote ops
\*/
#define gaPermuteProcList(nproc)\
{\
  if((nproc) ==1) ProcListPerm[0]=0; \
  else{\
    int _i, iswap, temp;\
    if((nproc) > GAnproc) ga_error("permute_proc: error ", (nproc));\
\
    /* every process generates different random sequence */\
    (void)srand((unsigned)GAme); \
\
    /* initialize list */\
    for(_i=0; _i< (nproc); _i++) ProcListPerm[_i]=_i;\
\
    /* list permutation generated by random swapping */\
    for(_i=0; _i< (nproc); _i++){ \
      iswap = (int)(rand() % (nproc));  \
      temp = ProcListPerm[iswap]; \
      ProcListPerm[iswap] = ProcListPerm[_i]; \
      ProcListPerm[_i] = temp; \
    } \
  }\
}
     

extern "C" {
  void ga_ma_base_address_(Integer*, Void**) ;
}

/*\ INITIALIZE GLOBAL ARRAY STRUCTURES
 *
 *  either ga_initialize_ltd or ga_initialize must be the first 
 *         GA routine called (except ga_uses_ma)
\*/
void FATR  ga_initialize_()
{
Integer  i;
#ifdef CHECK_MA_ALGN
Integer  off_dbl, off_int, off_dcpl;
#endif

    if(GAinitialized) return;

    /* zero in pointers in GA array */
    for(i=0;i<MAX_ARRAYS; i++)
       GA[i].ptr  = (char**)0;
    GAmaster= 0;
    GAnproc = (Integer)ga_msg_nnodes_();
    GAme = (Integer)ga_msg_nodeid_();
    MPme= ga_msg_nodeid_();
    MPnproc = ga_msg_nnodes_();

    if(GAnproc > MAX_NPROC && MPme==0){
      fprintf(stderr,"Current GA setup is for up to %d processors\n",MAX_NPROC);
      fprintf(stderr,"Please change MAX_NPROC in config.h & recompile\n");
      ga_error("terminating...",0);
    }

    map = (Integer*)malloc((GAnproc*2*MAXDIM +1)*sizeof(Integer));
    if(!map) ga_error("ga_init:malloc failed (map)",0);
    proclist = (Integer*)malloc(GAnproc*sizeof(Integer));
    if(!proclist) ga_error("ga_init:malloc failed (proclist)",0);

    /* set activity status for all arrays to inactive */
    for(i=0;i<max_global_array;i++)GA[i].actv=0;

    GAinitialized = 1;

    /* Initialize MA-like addressing:
     *    get addressees for the base arrays for double, complex and int types
     *
     * MA include files: macommon.h, macdecls.h and mafdecls.h
     */

    INT_MB = (Integer*)MA_get_mbase(MT_F_INT);
    DBL_MB = (DoublePrecision*)MA_get_mbase(MT_F_DBL);
    DCPL_MB= (DoubleComplex*)MA_get_mbase(MT_F_DCPL);

#   ifdef CHECK_MA_ALGN
        off_dbl = 0 != ((long)DBL_MB)%sizeof(DoublePrecision);
        off_int = 0 != ((long)INT_MB)%sizeof(Integer);
        off_dcpl= 0 != ((long)DCPL_MB)%sizeof(DoublePrecision);

        if(off_dbl)
           ga_error("ga_initialize: MA DBL_MB not alligned",
                    (Integer)(long)DBL_MB);

        if(off_int)
           ga_error("ga_initialize: INT_MB not alligned",
                    (Integer)(long)INT_MB);

        if(off_dcpl)
          ga_error("ga_initialize: DCPL_MB not alligned",
                    (Integer)(long)DCPL_MB);
#   endif

    if(DEBUG)
        printf("%d INT_MB=%d(%x) DBL_MB=%ld(%lx) DCPL_MB=%d(%lx)\n",
                GAme, INT_MB,INT_MB, DBL_MB,DBL_MB, DCPL_MB,DCPL_MB);
    ARMCI_Init();
    AdlibInit_() ;
}




/*\ IS MA USED FOR ALLOCATION OF GA MEMORY ?
\*/ 
logical FATR ga_uses_ma_()
{
#  if defined(CRAY_T3E)||defined(SP)||defined(LAPI)||defined(PARAGON)||defined(ARMCI)
     return FALSE;
#  else
     return TRUE;
#  endif
}


/*\ IS MEMORY LIMIT SET ?
\*/
logical FATR ga_memory_limited_()
{
   if(GA_memory_limited) return TRUE;
   else                  return FALSE;
}



/*\ RETURNS AMOUNT OF MEMORY on each processor IN ACTIVE GLOBAL ARRAYS 
\*/
Integer  FATR ga_inquire_memory_()
{
Integer i, sum=0;
    for(i=0; i<max_global_array; i++) 
        if(GA[i].actv) sum += GA[i].size; 
    return(sum);
}



/*\ RETURNS AMOUNT OF GA MEMORY AVAILABLE on calling processor 
\*/
Integer FATR ga_memory_avail_()
{
#ifdef SYSV
   return(GA_total_memory); 
#else
   Integer ma_limit = MA_inquire_avail(MT_F_BYTE);

   if ( GA_memory_limited ) return( MIN(GA_total_memory, ma_limit) );
   else return( ma_limit );
#endif
}


#define MBUFLEN 256
#define MBUF_LEN MBUFLEN+2
static double ga_int_malloc_buf[MBUF_LEN];
static int mbuf_used=0;
#define MBUF_GUARD -1998.1998
void *gai_malloc(int bytes)
{
    void *ptr;
    if(!mbuf_used && bytes <= MBUF_LEN){
       if(DEBUG){
          ga_int_malloc_buf[0]= MBUF_GUARD;
          ga_int_malloc_buf[MBUFLEN]= MBUF_GUARD;
       }
       ptr = ga_int_malloc_buf+1;
       mbuf_used++;
    }else{
        Integer handle, idx, elems = (bytes+sizeof(double)-1)/sizeof(double)+1; 
        if(MA_push_get(MT_DBL, elems, "temp", &handle, &idx)){
            MA_get_pointer(handle, (Pointer*) &ptr);
            *((Integer*)ptr)= handle;
            ptr = ((double*)ptr)+ 1;  /*needs sizeof(double)>=sizeof(Integer) */
        }else
            ptr=NULL;
    }
    return ptr;
}

void gai_free(void *ptr)
{
    if(ptr == (ga_int_malloc_buf+1)){
        if(DEBUG){
          assert(ga_int_malloc_buf[0]== MBUF_GUARD);
          assert(ga_int_malloc_buf[MBUFLEN]== MBUF_GUARD);
          assert(mbuf_used ==1);
        }
        mbuf_used =0;
    }else{
        Integer handle= *( (Integer*) (-1 + (double*)ptr));
        if(!MA_pop_stack(handle)) ga_error("gai_free:MA_pop_stack failed",0);
    }
}

        

/*\ INITIALIZE GLOBAL ARRAY STRUCTURES and SET LIMIT ON GA MEMORY USAGE
 *  
 *  the byte limit is per processor (even for shared memory)
 *  either ga_initialize_ltd or ga_initialize must be the first 
 *         GA routine called (except ga_uses_ma)
 *  ga_initialize_ltd is another version of ga_initialize 
 *         without memory control
 *  mem_limit < 0 means "memory unlimited"
\*/
void FATR  ga_initialize_ltd_(Integer* mem_limit)
{

  GA_total_memory = *mem_limit; 
  if(*mem_limit >= 0) GA_memory_limited = 1; 
  ga_initialize_();
}


#define gam_checktype(_type)\
       if(_type != MT_F_DBL && _type != MT_F_INT &&  _type != MT_F_DCPL)\
         ga_error("type not yet supported ",  _type)

#define gam_checkdim(ndim, dims)\
{\
int _d;\
    if(ndim<1||ndim>MAXDIM) ga_error("unsupported number of dimensions",ndim);\
    for(_d=0; _d<ndim; _d++)\
         if(dims[_d]<1)ga_error("wrong dimension specified",dims[_d]);\
}

void gai_init_struct(Integer handle)
{
     if(!GA[handle].ptr){
        int len = MIN(GAnproc, MAX_PTR);
        GA[handle].ptr = (char**)malloc(len*sizeof(char**));
     }
     if(!GA[handle].ptr)ga_error("malloc failed: ptr:",0);
}

/*\ get memory alligned w.r.t. MA base
 *  required on Linux as g77 ignores natural data alignment in common blocks
\*/ 
int gai_getmem(char **ptr_arr, int bytes, int type, long *adj)
{
int status;
#ifndef _CHECK_MA_ALGN
char *base;
long diff, item_size;  
Integer *adjust;
int i;

    /* need to enforce proper, natural allignment (on size boundary)  */
    switch (type){
      case MT_F_DBL: base =  (char *) DBL_MB; break;
      case MT_F_INT: base =  (char *) INT_MB; break;
      case MT_F_DCPL: base =  (char *) DCPL_MB; break;
      default:        base = (char*)0;
    }

    item_size = GAsizeofM(type);
    bytes += item_size; /***** will change on clusters *****/

#endif

    *adj = 0;
    status = ARMCI_Malloc((void**)ptr_arr, bytes);
/*    printf("%d ptr=%d\n",GAme,ptr_arr[GAme]);*/
    if(status) return status;

#ifndef _CHECK_MA_ALGN

    /* adjust all addresses if they are not alligned on corresponding nodes*/
    adjust = (Integer*)gai_malloc(GAnproc*sizeof(Integer));

    diff = (ABS( base - (char *) ptr_arr[GAme])) % item_size; 
    for(i=0;i<GAnproc;i++)adjust[i]=0;
    adjust[GAme] = (diff > 0) ? item_size - diff : 0;
    *adj = adjust[GAme];

    ga_igop(GA_TYPE_GSM, adjust, GAnproc, "+");
    
    for(i=0;i<GAnproc;i++){
/*       printf("%d: adjust[%d]=%d\n",GAme, i, adjust[i]);*/
       ptr_arr[i] = adjust[i] + (char*)ptr_arr[i];
    }

#endif
    return status;
}




/*\ CREATE A GLOBAL ARRAY -- IRREGULAR DISTRIBUTION
\*/
logical nga_create_irreg(
        Integer type,    /* MA type */ 
        Integer ndim,    /* number of dimensions */
        Integer dims[],   /* array of dimensions */
        char *array_name, /* array name */
        Integer map[],    /* decomposition map array */ 
        Integer nblock[], /* number of blocks for each dimension in map */
        Integer *g_a)     /* array handle (output) */
{

Integer  mem_size, nelem, mem_size_proc;
Integer  i, ga_handle, status ;

      ga_sync_();
      GA_PUSH_NAME("nga_create_irreg");

      if(!GAinitialized) ga_error("GA not initialized ", 0);
      gam_checktype(type);
      gam_checkdim(ndim, dims);

      GAstat.numcre ++;

      /*** Get next free global array handle ***/
      ga_handle =-1; i=0;
      do{
          if(!GA[i].actv) ga_handle=i;
          i++;
      }while(i<max_global_array && ga_handle==-1);
      if( ga_handle == -1)
          ga_error("ga_create: too many arrays ", (Integer)max_global_array);
      *g_a = (Integer)ga_handle - GA_OFFSET;

      /*** fill in Global Info Record for g_a ***/
      gai_init_struct(ga_handle);
      GA[ga_handle].type = type;
      GA[ga_handle].actv = 1;
      strcpy(GA[ga_handle].name, array_name);

      int nblocki [MAXDIM] ;
      for(i = 0 ; i < ndim ; i++)
        nblocki [i] = nblock [i] ;
      Procs* p = getProcs(ndim, nblocki) ;
      nelem = 1 ;
      Map* maps = GA[ga_handle].maps ;
      for(i = 0; i < ndim; i++) {
         IrregRange x(dims [i], p->dim(i), map) ;

         maps [i] = Map(x, nelem) ;
         nelem   *= x.volume() ;

         map += nblock [i] ;
      }
      GA[ga_handle].dad = DAD(ndim, *p, maps) ;

      GA[ga_handle].elemsize = GAsizeofM(type);

      mem_size = mem_size_proc =  nelem * GA[ga_handle].elemsize;
      GA[ga_handle].id = INVALID_MA_HANDLE;
      GA[ga_handle].size = mem_size_proc;


      /* if requested, enforce limits on memory consumption */
      if(GA_memory_limited) GA_total_memory -= mem_size_proc;

      /* check if everybody has enough memory left */
      if(GA_memory_limited){
         status = (GA_total_memory >= 0) ? 1 : 0;
         ga_igop(GA_TYPE_GSM, &status, 1, "*");
      }else status = 1;

/*      fprintf(stderr,"%d, elems=%d size=%d status=%d\n",GAme,nelem,mem_size,status);*/
/*      ga_sync_();*/
      if(status){
          status = !gai_getmem(GA[ga_handle].ptr,mem_size,(int)type,
                               &GA[ga_handle].id);
      }else{
          GA[ga_handle].ptr[GAme]=NULL;
      }

/*      fprintf(stderr,"%d, status=%d\n",GAme,status);*/
      ga_sync_();

      if(status){
         GAstat.curmem += GA[ga_handle].size;
         GAstat.maxmem  = MAX(GAstat.maxmem, GAstat.curmem);
         status = TRUE;
      }else{
         ga_destroy_(g_a);
         status = FALSE;
      }

      GA_POP_NAME;
      return status;
}



/*\ print subscript of ndim dimensional array with two strings before and after
\*/
static void print_subscript(char *pre,int ndim, Integer subscript[], char* post)
{
        int i;

        printf("%s [",pre);
        for(i=0;i<ndim;i++){
                printf("%d",subscript[i]);
                if(i==ndim-1)printf("] %s",post);
                else printf(",");
        }
}


/*\ CREATE A GLOBAL ARRAY -- IRREGULAR DISTRIBUTION
\*/
logical ga_create_irreg(Integer* type, Integer* dim1, Integer* dim2,
                        char* array_name,
                        Integer* map1, Integer* nblock1,
                        Integer* map2, Integer* nblock2,
                        Integer* g_a)
     /*
      * array_name    - a unique character string [input]
      * type          - MA type [input]
      * dim1/2        - array(dim1,dim2) as in FORTRAN [input]
      * nblock1       - no. of blocks first dimension is divided into [input]
      * nblock2       - no. of blocks second dimension is divided into [input]
      * map1          - no. ilo in each block [input]
      * map2          - no. jlo in each block [input]
      * g_a           - Integer handle for future references [output]
      */
{
Integer  mem_size, nelem, mem_size_proc;
Integer  i, ga_handle, status;

      if(!GAinitialized) ga_error("GA not initialized ", 0);

      ga_sync_();

      GAstat.numcre ++; 

      if(*type != MT_F_DBL && *type != MT_F_INT &&  *type != MT_F_DCPL)
         ga_error("ga_create_irreg: type not yet supported ",  *type);
      else if( *dim1 <= 0 )
         ga_error("ga_create_irreg: array dimension1 invalid ",  *dim1);
      else if( *dim2 <= 0)
         ga_error("ga_create_irreg: array dimension2 invalid ",  *dim2);
      else if(*nblock1 <= 0)
         ga_error("ga_create_irreg: nblock1 <=0  ",  *nblock1);
      else if(*nblock2 <= 0)
         ga_error("ga_create_irreg: nblock2 <=0  ",  *nblock2);
      else if(*nblock1 * *nblock2 > GAnproc)
         ga_error("ga_create_irreg: too many blocks ",*nblock1 * *nblock2);

      if(GAme==0&& DEBUG){
        fprintf(stderr," array:%d map1:\n", *g_a);
        for (i=0;i<*nblock1;i++)fprintf(stderr," %d |",map1[i]);
        fprintf(stderr," \n array:%d map2:\n", *g_a);
        for (i=0;i<*nblock2;i++)fprintf(stderr," %d |",map2[i]);
        fprintf(stderr,"\n\n");
      }

      /*** Get next free global array handle ***/
      ga_handle =-1; i=0;
      do{
          if(!GA[i].actv) ga_handle=i;
          i++;
      }while(i<max_global_array && ga_handle==-1);
      if( ga_handle == -1)
          ga_error("ga_create: too many arrays ", (Integer)max_global_array);
      *g_a = (Integer)ga_handle - GA_OFFSET;

      /*** fill in Global Info Record for g_a ***/
      gai_init_struct(ga_handle);
      GA[ga_handle].type = *type;
      GA[ga_handle].actv = 1;
      strcpy(GA[ga_handle].name, array_name);

      int nblock [2] ;
      nblock [0] = *nblock1 ;
      nblock [1] = *nblock2 ;
      Procs* p = getProcs(2, nblock) ;

      IrregRange x(*dim1, p->dim(i), map1) ;
      IrregRange y(*dim2, p->dim(i), map2) ;

      Map* maps = GA[ga_handle].maps ;
      maps [0] = Map(x, 1) ;
      nelem    = x.volume() ;
      maps [1] = Map(y, nelem) ;
      nelem   *= y.volume() ;

      GA[ga_handle].dad = DAD(2, *p, maps) ;

      GA[ga_handle].elemsize = GAsizeofM(*type);

      /*
      if(GAme ==0 && DEBUG){
         fprintf(stderr,"\nmapc %d elem\n", *nblock1 + *nblock2);
         for(i=0;i<1+*nblock1+ *nblock2;i++)
             fprintf(stderr,"%d,",GA[ga_handle].mapc[i]);
         fprintf(stderr,"\n\n");
      }
      */

      mem_size = mem_size_proc =  nelem * GA[ga_handle].elemsize;
      GA[ga_handle].id = INVALID_MA_HANDLE;

      /* on shmem platforms, we use avg mem_size per processor in cluster */
      GA[ga_handle].size = mem_size_proc;

      /* if requested, enforce limits on memory consumption */
      if(GA_memory_limited) GA_total_memory -= mem_size_proc; 


      /* check if everybody has enough memory left */
      if(GA_memory_limited){
         status = (GA_total_memory >= 0) ? 1 : 0;
         ga_igop(GA_TYPE_GSM, &status, 1, "*");
      }else status = 1;
 
      if(status){
          status = !gai_getmem(GA[ga_handle].ptr,mem_size,*type,
                               &GA[ga_handle].id);
      }else{
          GA[ga_handle].ptr[GAme]=NULL;
      }

      ga_sync_();

#     ifdef GA_CREATE_INDEF
      if(status){
         Integer one = 1;
         if (GAme == 0) fprintf(stderr,"Initializing GA array%ld\n",*g_a);
         if(*type == MT_F_DBL){ 
             double bad = DBL_MAX;
             ga_fill_patch_(g_a, &one, dim1, &one, dim2, (Void *) &bad);
         } else if (*type == MT_F_INT) {
             Integer bad = (Integer) INT_MAX;
             ga_fill_patch_(g_a, &one, dim1, &one, dim2, (Void *) &bad);
         } else if (*type == MT_F_DCPL) {
             DoubleComplex bad = {DBL_MAX, DBL_MAX};
             ga_fill_patch_(g_a, &one, dim1, &one, dim2, (Void *) &bad);
         } else {
             ga_error("ga_create_irreg: type not yet supported ",  *type);
         }
      }
#     endif

      if(status){
         GAstat.curmem += GA[ga_handle].size;
         GAstat.maxmem  = MAX(GAstat.maxmem, GAstat.curmem);
         return(TRUE);
      }else{
         ga_destroy_(g_a);
         return(FALSE);
      }
}


/*\ CREATE A GLOBAL ARRAY -- IRREGULAR DISTRIBUTION
 *  Fortran version
\*/
#if defined(CRAY) || defined(WIN32)
logical ga_create_irreg_(Integer* type, Integer* dim1, Integer* dim2,
                         _fcd array_name,
                         Integer* map1, Integer* nblock1,
                         Integer* map2, Integer* nblock2,
                         Integer* g_a)
#else
logical ga_create_irreg_(Integer* type, Integer* dim1, Integer* dim2,
                         char* array_name,
                         Integer* map1, Integer* nblock1,
                         Integer* map2, Integer* nblock2,
                         Integer* g_a, int slen)
#endif
{
char buf[FNAM];
#if defined(CRAY) || defined(WIN32)
      f2cstring(_fcdtocp(array_name), _fcdlen(array_name), buf, FNAM);
#else
      f2cstring(array_name ,slen, buf, FNAM);
#endif
  return( ga_create_irreg(type, dim1, dim2, buf, map1, nblock1,
                         map2, nblock2, g_a));
}


Integer mapALL[MAX_NPROC+1];

extern "C" {
  extern void ddb(Integer ndims, Integer dims[], Integer npes, Integer blk[], Integer pedims[]);
}

logical nga_create(Integer type,
                   Integer ndim,
                   Integer dims[],
                   char* array_name,
                   Integer chunk[],
                   Integer *g_a)
{
Integer pe[MAXDIM], *pmap[MAXDIM], *map;
Integer d, i;
Integer blk[MAXDIM];
logical status;

      GA_PUSH_NAME("nga_create");
      if(!GAinitialized) ga_error("GA not initialized ", 0);
      gam_checktype(type);
      gam_checkdim(ndim, dims);

      if(chunk && chunk[0]!=0) /* for either NULL or chunk[0]=0 compute all */
          for(d=0; d< ndim; d++) blk[d]=chunk[d];
      else
          for(d=0; d< ndim; d++) blk[d]=-1;

      ddb(ndim, dims, GAnproc, blk, pe);
/*      pe[0]=pe[1]=1; pe[2]=8;*/
/*      blk[0]=blk[1]=8; blk[2]=1;*/

      for(d=0, map=mapALL; d< ndim; d++){
         Integer nblock;
         pmap[d] = map;
         for(i=0,nblock=0; i< dims[d]; i += blk[d], nblock++) map[nblock]=i+1;
         pe[d] = MIN(pe[d],nblock);
         map +=  pe[d]; 
      }

      if(GAme==0&& DEBUG){
         print_subscript("pe ",ndim, pe,"\n");
         print_subscript("blocks ",ndim, blk,"\n");
         printf("decomposition map\n");
         for(d=0; d< ndim; d++){
           printf("dim=%d: ",d); 
           for (i=0;i<pe[d];i++)printf("%d ",pmap[d][i]);
           printf("\n"); 
         }
         fflush(stdout);
      }

      status = nga_create_irreg(type, ndim, dims, array_name, mapALL, pe, g_a);

/*
      status = ga_create_irreg( type, &dims[0], &dims[1], array_name, pmap[0], 
              &pe[0], pmap[1], &pe[1], g_a);
*/

      GA_POP_NAME;
      return status;
}


      

logical ga_create(Integer* type, Integer* dim1, Integer* dim2,
                  char* array_name, Integer* chunk1, Integer* chunk2,
                  Integer* g_a)
{
Integer ndim=2, dims[2], chunk[2];

    dims[0]=*dim1;
    dims[1]=*dim2;
    chunk[0]=*chunk1;
    chunk[1]=*chunk2;
        
    return nga_create(*type, ndim,  dims, array_name, chunk, g_a); 
}

/*\ CREATE A GLOBAL ARRAY
\*/
logical oga_create(Integer* type, Integer* dim1, Integer* dim2, 
                   char* array_name, Integer* chunk1, Integer* chunk2,
                   Integer* g_a)
     /*
      * array_name    - a unique character string [input]
      * type          - MA type [input]
      * dim1/2        - array(dim1,dim2) as in FORTRAN [input]
      * chunk1/2      - minimum size that dimensions should
      *                 be chunked up into [input]
      *                 setting chunk1=dim1 gives distribution by rows
      *                 setting chunk2=dim2 gives distribution by columns 
      *                 Actual chunk sizes are modified so that they are
      *                 at least the min size and each process has either
      *                 zero or one chunk. 
      *                 chunk1/2 <=1 yields even distribution
      * g_a           - Integer handle for future references [output]
      */
{
int   i, nprocx, nprocy, fchunk1, fchunk2;
static Integer map1[MAX_NPROC], map2[MAX_NPROC];
Integer nblock1, nblock2;

      if(!GAinitialized) ga_error("GA not initialized ", 0);

      /* sync is in ga_create_irreg */

      if(*type != MT_F_DBL && *type != MT_F_INT &&  *type != MT_F_DCPL)
         ga_error("ga_create: type not yet supported ",  *type);
      else if( *dim1 <= 0 )
         ga_error("ga_create: array dimension1 invalid ",  *dim1);
      else if( *dim2 <= 0)
         ga_error("ga_create: array dimension2 invalid ",  *dim2);

      /*** figure out chunking ***/
      if(*chunk1 <= 1 && *chunk2 <= 1){
        if(*dim1 == 1)      { nprocx =1; nprocy=(int)GAnproc;}
        else if(*dim2 == 1) { nprocy =1; nprocx=(int)GAnproc;}
        else {
           /* nprocx= (int)sqrt((double)GAnproc);*/
           double dproc = ((double)GAnproc*(*dim1))/((double) *dim2);
           nprocx = NEAR_INT(sqrt(dproc)); 
           nprocx = MAX(1, nprocx); /* to avoid division by 0 */
           for(i=nprocx;i>0&& (GAnproc%i);i--);
           nprocx =(int)i; nprocy=(int)GAnproc/nprocx;
        }

        fchunk1 = (int) MAX(1, *dim1/nprocx);
        fchunk2 = (int) MAX(1, *dim2/nprocy);
      }else if(*chunk1 <= 1){
        fchunk1 = (int) MAX(1, (*dim1 * *dim2)/(GAnproc* *chunk2));
        fchunk2 = (int) *chunk2;
      }else if(*chunk2 <= 1){
        fchunk1 = (int) *chunk1;
        fchunk2 = (int) MAX(1, (*dim1 * *dim2)/(GAnproc* *chunk1));
      }else{
        fchunk1 = (int) MAX(1,  *chunk1);
        fchunk2 = (int) MAX(1,  *chunk2);
      }

      fchunk1 = (int)MIN(fchunk1, *dim1);
      fchunk2 = (int)MIN(fchunk2, *dim2);

      /*** chunk size correction for load balancing ***/
      while(((*dim1-1)/fchunk1+1)*((*dim2-1)/fchunk2+1) >GAnproc){
           if(fchunk1 == *dim1 && fchunk2 == *dim2) 
                     ga_error("ga_create: chunking failed !! ", 0L);
           if(fchunk1 < *dim1) fchunk1 ++; 
           if(fchunk2 < *dim2) fchunk2 ++; 
      }

      /* Now build map arrays */
      for(i=0, nblock1=0; i< *dim1; i += fchunk1, nblock1++) map1[nblock1]=i+1;
      for(i=0, nblock2=0; i< *dim2; i += fchunk2, nblock2++) map2[nblock2]=i+1;   
      if(GAme==0&& DEBUG){
         fprintf(stderr,"blocks (%d,%d)\n",nblock1, nblock2);
         fprintf(stderr,"chunks (%d,%d)\n",fchunk1, fchunk2);
         if(GAme==0){
           for (i=0;i<nblock1;i++)fprintf(stderr," %d ",map1[i]);
           for (i=0;i<nblock2;i++)fprintf(stderr," .%d ",map2[i]);
         }
      }

      return( ga_create_irreg(type, dim1, dim2, array_name, map1, &nblock1,
                         map2, &nblock2, g_a) );
}



/*\ CREATE A GLOBAL ARRAY
 *  Fortran version
\*/
#if defined(CRAY) || defined(WIN32)
logical FATR ga_create_(Integer* type, Integer* dim1, Integer* dim2,
                        _fcd array_name, Integer* chunk1, Integer* chunk2,
                        Integer* g_a)
#else
logical ga_create_(Integer* type, Integer* dim1, Integer* dim2,
                   char* array_name, Integer* chunk1, Integer* chunk2,
                   Integer* g_a, int slen)
#endif
{
char buf[FNAM];
#if defined(CRAY) || defined(WIN32)
      f2cstring(_fcdtocp(array_name), _fcdlen(array_name), buf, FNAM);
#else
      f2cstring(array_name ,slen, buf, FNAM);
#endif

  return(ga_create(type, dim1, dim2, buf, chunk1, chunk2, g_a));
}


/*\ CREATE A GLOBAL ARRAY
 *  Fortran version
\*/
#if defined(CRAY) || defined(WIN32)
logical FATR nga_create_(Integer *type, Integer *ndim, Integer *dims,
                   _fcd array_name, Integer *chunk, Integer *g_a)
#else
logical FATR nga_create_(Integer *type, Integer *ndim, Integer *dims,
                   char* array_name, Integer *chunk, Integer *g_a, int slen)
#endif
{
char buf[FNAM];
#if defined(CRAY) || defined(WIN32)
      f2cstring(_fcdtocp(array_name), _fcdlen(array_name), buf, FNAM);
#else
      f2cstring(array_name ,slen, buf, FNAM);
#endif

  return (nga_create(*type, *ndim,  dims, buf, chunk, g_a));
}



/*\ RETURN COORDINATES OF A GA PATCH ASSOCIATED WITH PROCESSOR proc
\*/
void FATR nga_distribution_(Integer *g_a, Integer *proc, Integer *lo, Integer *
hi)
{
Integer ga_handle;

   ga_check_handleM(g_a, "aga_distribution");
   ga_handle = (GA_OFFSET + *g_a);
   ga_ownsM(ga_handle, *proc, lo, hi);
}


Integer ga_ndim_(Integer *g_a)
{
      ga_check_handleM(g_a,"ga_ndim");
      return GA[*g_a +GA_OFFSET].dad.rnk();
}



/*\ DUPLICATE A GLOBAL ARRAY
 *  -- new array g_b will have properties of g_a
\*/
logical ga_duplicate(Integer* g_a, Integer* g_b, char* array_name)
     /*
      * array_name    - a character string [input]
      * g_a           - Integer handle for reference array [input]
      * g_b           - Integer handle for new array [output]
      */
{
char     **save_ptr;
Integer  mem_size, mem_size_proc;
Integer  i, ga_handle, status;

      ga_sync_();

      GAstat.numcre ++; 

      ga_check_handleM(g_a,"ga_duplicate");       

      /* find a free global_array handle for g_b */
      ga_handle =-1; i=0;
      do{
        if(!GA[i].actv) ga_handle=i;
        i++;
      }while(i<max_global_array && ga_handle==-1);
      if( ga_handle == -1)
          ga_error("ga_duplicate: too many arrays ", (Integer)max_global_array);
      *g_b = (Integer)ga_handle - GA_OFFSET;

      gai_init_struct(ga_handle);

      /*** copy content of the data structure ***/
      save_ptr = GA[ga_handle].ptr;
      GA[ga_handle] = GA[GA_OFFSET + *g_a];
      strcpy(GA[ga_handle].name, array_name);
      GA[ga_handle].ptr = save_ptr;

      /*** Memory Allocation & Initialization of GA Addressing Space ***/
      mem_size = mem_size_proc = GA[ga_handle].size; 
      GA[ga_handle].id = INVALID_MA_HANDLE;
      /* if requested, enforce limits on memory consumption */
      if(GA_memory_limited) GA_total_memory -= mem_size_proc;

     /* check if everybody has enough memory left */
      if(GA_memory_limited){
         status = (GA_total_memory >= 0) ? 1 : 0;
         ga_igop(GA_TYPE_GSM, &status, 1, "*");
      }else status = 1;

      if(status)
          status = !gai_getmem(GA[ga_handle].ptr,mem_size,
                               GA[ga_handle].type, &GA[ga_handle].id);
      else{
          GA[ga_handle].ptr[GAme]=NULL;
      }

      ga_sync_();

#     ifdef GA_CREATE_INDEF
      if(status){
         Integer one = 1; 
         Integer dim1 =GA[ga_handle].dims[1], dim2=GA[ga_handle].dims[2];
         if(GAme==0)fprintf(stderr,"duplicate:initializing GA array%ld\n",*g_b);
         if(GA[ga_handle].type == MT_F_DBL) {
             DoublePrecision bad = DBL_MAX;
             ga_fill_patch_(g_b, &one, &dim1, &one, &dim2,  &bad);
         } else if (GA[ga_handle].type == MT_F_INT) {
             Integer bad = (Integer) INT_MAX;
             ga_fill_patch_(g_b, &one, &dim1, &one, &dim2,  &bad);
         } else if (GA[ga_handle].type == MT_F_DCPL) {
             DoubleComplex bad = {DBL_MAX, DBL_MAX};
             ga_fill_patch_(g_b, &one, &dim1, &one, &dim2,  &bad);
         } else {
             ga_error("ga_duplicate: type not supported ",GA[ga_handle].type);
         }
      }
#     endif

      if(status){
         GAstat.curmem += GA[ga_handle].size;
         GAstat.maxmem  = MAX(GAstat.maxmem, GAstat.curmem);
         return(TRUE);
      }else{ 
         ga_destroy_(g_b);
         return(FALSE);
      }
}


/*\ DUPLICATE A GLOBAL ARRAY
 *  Fortran version
\*/
#if defined(CRAY) || defined(WIN32)
logical FATR ga_duplicate_(Integer* g_a, Integer* g_b, _fcd array_name)
#else
logical FATR ga_duplicate_(Integer* g_a, Integer* g_b,
                           char* array_name, int slen)
#endif
{
char buf[FNAM];
#if defined(CRAY) || defined(WIN32)
      f2cstring(_fcdtocp(array_name), _fcdlen(array_name), buf, FNAM);
#else
      f2cstring(array_name ,slen, buf, FNAM);
#endif

  return(ga_duplicate(g_a, g_b, buf));
}



/*\ DESTROY A GLOBAL ARRAY
\*/
logical FATR ga_destroy_(Integer *g_a)
{
Integer ga_handle = GA_OFFSET + *g_a;

    ga_sync_();
    GAstat.numdes ++; /*regardless of array status we count this call */

    /* fails if handle is out of range or array not active */
    if(ga_handle < 0 || ga_handle >= max_global_array) return FALSE;
    if(GA[ga_handle].actv==0) return FALSE;       
    if(GA[ga_handle].ptr[GAme]==NULL) return TRUE;
 
    /* make sure that we free original (before address allignment) pointer */
    ARMCI_Free(GA[ga_handle].ptr[GAme] - GA[ga_handle].id);
/*    printf("%d destr ptr=%d\n",GAme,GA[ga_handle].ptr[GAme]);*/
    if(GA_memory_limited) GA_total_memory += GA[ga_handle].size;
    GA[ga_handle].actv = 0;     
    GAstat.curmem -= GA[ga_handle].size;

    return(TRUE);
}

    
     
/*\ TERMINATE GLOBAL ARRAY STRUCTURES
 *
 *  all GA arrays are destroyed & shared memory is dealocated
 *  GA routines (except for ga_initialize) should not be called thereafter 
\*/
void FATR  ga_terminate_() 
{
Integer i, handle;
extern double t_dgop, n_dgop, s_dgop;


    if(!GAinitialized) return;
    for (i=0;i<max_global_array;i++){
          handle = i - GA_OFFSET ;
          if(GA[i].actv) ga_destroy_(&handle);
          if(GA[i].ptr) free(GA[i].ptr);
    }
    ga_sync_();

    GA_total_memory = -1; /* restore "unlimited" memory usage status */
    GA_memory_limited = 0;
    free(map);
    free(proclist);

    ARMCI_Finalize();
    AdlibFinalize_() ;
    GAinitialized = 0;
}   

    
/*\ IS ARRAY ACTIVE/INACTIVE
\*/ 
Integer FATR ga_verify_handle_(Integer* g_a)
{
  return (Integer)
    ((*g_a + GA_OFFSET>= 0) && (*g_a + GA_OFFSET< max_global_array) && 
             GA[GA_OFFSET + (*g_a)].actv);
}
 


#define gaShmemLocation(proc, g_a, _i, _j, ptr_loc, ld)                        \
{                                                                              \
Integer _ilo, _ihi, _jlo, _jhi, offset, proc_place, g_handle=(g_a)+GA_OFFSET;  \
Integer _lo[2], _hi[2];                                                        \
                                                                               \
      ga_ownsM(g_handle, (proc),_lo,_hi);                                      \
      _ilo = _lo[0]; _ihi=_hi[0];                                              \
      _jlo = _lo[1]; _jhi=_hi[1];                                              \
                                                                               \
      if((_i)<_ilo || (_i)>_ihi || (_j)<_jlo || (_j)>_jhi){                    \
          sprintf(err_string,"%s: p=%d invalid i/j (%d,%d) >< (%d:%d,%d:%d)",  \
                 "gaShmemLocation", proc, (_i),(_j), _ilo, _ihi, _jlo, _jhi);  \
          ga_error(err_string, g_a );                                          \
      }                                                                        \
      offset = ((_i) - _ilo) + (_ihi-_ilo+1)*((_j)-_jlo);                      \
                                                                               \
      /* find location of the proc in current cluster pointer array */         \
      proc_place =  proc;                                             \
      *(ptr_loc) = GA[g_handle].ptr[proc_place] +                              \
                   offset*GAsizeofM(GA[g_handle].type);                        \
      *(ld) = _ihi-_ilo+1;                                                     \
}


#define gaCheckSubscriptM(subscr, lo, hi, ndim)                             \
{                                                                              \
Integer _d;                                                                    \
   for(_d=0; _d<  ndim; _d++)                                                  \
      if( subscr[_d]<  lo[_d] ||  subscr[_d]>  hi[_d]){                  \
          sprintf(err_string,"check subscript failed: %d not in (%d:%d) dim=", \
                  subscr[_d],  lo[_d],  hi[_d]);                            \
          ga_error(err_string, _d);                                            \
      }\
}


#define gam_Location(proc, g_handle,  subscript, ptr_loc, ld)                  \
{                                                                              \
Integer _offset=0, _d, _factor=1, _last=GA[g_handle].dad.rnk()-1;              \
Integer _lo[MAXDIM], _hi[MAXDIM];                                              \
                                                                               \
      ga_ownsM(g_handle, proc, _lo, _hi);                                      \
      gaCheckSubscriptM(subscript, _lo, _hi, GA[g_handle].dad.rnk());          \
      for(_d=0; _d < _last; _d++)            {                                 \
          _offset += (subscript[_d]-_lo[_d]) * _factor;                        \
          ld[_d] = _hi[_d] - _lo[_d]+1;                                        \
          _factor *= ld[_d];                                                   \
      }                                                                        \
      _offset += (subscript[_last]-_lo[_last]) * _factor;                      \
      *(ptr_loc) =  GA[g_handle].ptr[proc]+_offset*GA[g_handle].elemsize;      \
}




#define ga_check_regionM(g_a, ilo, ihi, jlo, jhi, string){                     \
   DAD* dad = &(GA[GA_OFFSET + *(g_a)].dad) ;                                  \
   if (*(ilo) <= 0 || *(ihi) > dad->rng(0).size() ||                           \
       *(jlo) <= 0 || *(jhi) > dad->rng(1).size() ||                           \
       *(ihi) < *(ilo) ||  *(jhi) < *(jlo)){                                   \
       sprintf(err_string,"%s:request(%d:%d,%d:%d) out of range (1:%d,1:%d)",  \
               string, *(ilo), *(ihi), *(jlo), *(jhi),                         \
               dad->rng(0).size(), dad->rng(1).size());                        \
       ga_error(err_string, *(g_a));                                           \
   }                                                                           \
}



/*\ local put of a 2-dimensional patch of data into a global array
\*/
void ga_put_local(Integer g_a, Integer ilo, Integer ihi,
                               Integer jlo, Integer jhi,
                  void* buf, Integer offset, Integer ld_loc, Integer proc)
{
char     *ptr_glob, *ptr_loc;
int      ld_glob, rows, cols, size=GAsizeofM(GA[GA_OFFSET + g_a].type);
int      count[2];

   GA_PUSH_NAME("ga_put_local");

   rows = ihi - ilo +1;
   cols = jhi - jlo +1;

   gaShmemLocation(proc, g_a, ilo, jlo, &ptr_glob, &ld_glob);
   ptr_loc = (char *)buf  +  size* offset;
   count[0] = rows*size; count[1]=cols;
   ld_loc *= size; ld_glob *= size;

   int ld_loci = ld_loc ;
   ARMCI_PutS(ptr_loc, &ld_loci, ptr_glob, &ld_glob, count, 1, proc); 

   GA_POP_NAME;
}

#define ga_RegionError(ndim, lo, hi, val){\
Integer _d, _l;\
   char *str= "cannot locate region: ";\
   sprintf(err_string, str); \
   _l = strlen(str);\
   for(_d=0; _d< ndim; _d++){ \
        sprintf(err_string+_l, "%d:%d ",lo[_d],hi[_d]);\
        _l=strlen(err_string);\
   }\
   ga_error(err_string, val);\
}

/*\fill array with value
\*/
void FATR ga_fill_(Integer *g_a, void* val)
{
int i,elems,handle=GA_OFFSET + *g_a;
char *ptr;

   GA_PUSH_NAME("ga_fill");
   ga_sync_();

   ga_check_handleM(g_a, "ga_fill");
   gam_checktype(GA[handle].type);
   elems = GA[handle].size/GA[handle].elemsize;
   ptr = GA[handle].ptr[GAme];

   switch (GA[handle].type){
   case MT_F_DCPL: elems*=2; /* fall through */ 
   case MT_F_DBL:  
        for(i=0; i<elems;i++)((DoublePrecision*)ptr)[i]=*(DoublePrecision*)val;
        break;
   case MT_F_INT:  
        for(i=0; i<elems;i++)((Integer*)ptr)[i]=*(Integer*)val;
        break;
   default:
        ga_error("type not supported",GA[handle].type);
   }
   ga_sync_();
   GA_POP_NAME;
}



/*\ PUT A 2-DIMENSIONAL PATCH OF DATA INTO A GLOBAL ARRAY
\*/
void FATR nga_put_(Integer *g_a, 
                   Integer *lo,
                   Integer *hi,
                   Void    *buf,
                   Integer *ld)
{
Integer  p, handle=GA_OFFSET + *g_a;
Integer  idx, ndim, size ;

      GA_PUSH_NAME("nga_put");

      size = GA[handle].elemsize;
      //type = GA[handle].type;

      DAD* dad = &(GA[handle].dad) ;
      ndim = dad->rnk() ;

      GAstat.numput++;

      for(int d = 0; d < ndim ; d++)
        if(lo [d] < 1 || hi[d] > dad->rng(d).size())
          ga_RegionError(ndim, lo, hi, *g_a);

      PatchTransfers i(ld, buf, dad, (void**) GA[handle].ptr, size, lo, hi) ;

//      i.loop(0, 1, 0, 0, 1, 1, dad->grp(), 0) ;
      i.loop(0, 1, 0, 0, 1, 1, dad->grp().lead_prc(), 0) ;
                                    // enumerate block transfers.

      gaPermuteProcList(i.numBlk);

      int*   blocks  = i.getBlks() ;
      void** locPtrs = i.getLocs() ;
      void** remPtrs = i.getRems() ;

      for(idx=0; idx < i.numBlk; idx++) {
          p = (Integer)ProcListPerm [idx] ;
	  int*  block  = blocks + i.blockLen * p ;
	  void* locPtr = locPtrs [p] ;
	  void* remPtr = remPtrs [p] ;

          const int remId        = block [MBLKT_REM] ;
          const int numStrLevels = block [MBLKT_NST] ;

          int* exts    = block + MBLKT_CTG ;  // `contigLen' prepended
                                              // to vector.
          int* locStrs = exts + numStrLevels + 1 ;
          int* remStrs = locStrs + numStrLevels ;

          ARMCI_PutS(locPtr, locStrs, remPtr, remStrs, exts,
                     numStrLevels, remId) ;

          GAbytes.puttot += block [MBLKT_SIZ] ;
          if(remId == GAme)
            GAbytes.putloc += block [MBLKT_SIZ] ;
      }

      GA_POP_NAME;
}



void FATR  ga_put_(Integer* g_a, Integer* ilo, Integer* ihi,
                   Integer* jlo, Integer* jhi, Void* buf, Integer* ld)
{
Integer lo[2], hi[2];

#ifdef GA_TRACE
   trace_stime_();
#endif

   lo[0]=*ilo;
   lo[1]=*jlo;
   hi[0]=*ihi;
   hi[1]=*jhi;
   nga_put_(g_a, lo, hi, buf, ld);

#ifdef GA_TRACE
   trace_etime_();
   op_code = GA_OP_PUT; 
   trace_genrec_(g_a, ilo, ihi, jlo, jhi, &op_code);
#endif
}



/*\ local get of a 2-dimensional patch of data from a global array
\*/
void ga_get_local(Integer g_a, Integer ilo, Integer ihi, Integer jlo,
                  Integer jhi, void* buf, Integer offset, Integer ld_loc,
                  Integer proc)
{
char     *ptr_glob, *ptr_loc;
int      ld_glob, rows, cols, size=GAsizeofM(GA[GA_OFFSET + g_a].type);
int      count[2];

   GA_PUSH_NAME("ga_get_local");

   rows = ihi - ilo +1;
   cols = jhi - jlo +1;

   gaShmemLocation(proc, g_a, ilo, jlo, &ptr_glob, &ld_glob);

   ptr_loc = (char *)buf  + size * offset;
   count[0] = rows*size; count[1]=cols;
   ld_loc *= size; ld_glob *= size;

   int ld_loci = ld_loc ;
   ARMCI_GetS(ptr_glob, &ld_glob, ptr_loc, &ld_loci, count, 1, proc); 

   GA_POP_NAME;
}



/*\ GET AN N-DIMENSIONAL PATCH OF DATA FROM A GLOBAL ARRAY
\*/
void FATR nga_get_(Integer *g_a,
                   Integer *lo,
                   Integer *hi,
                   Void    *buf,
                   Integer *ld)
{
Integer  p, handle=GA_OFFSET + *g_a;
Integer  idx, ndim, size ;

      GA_PUSH_NAME("nga_get");

      size = GA[handle].elemsize;
      //type = GA[handle].type;

      DAD* dad = &(GA[handle].dad) ;
      ndim = dad->rnk() ;

      GAstat.numget++;

      for(int d = 0; d < ndim ; d++)
        if(lo [d] < 1 || hi[d] > dad->rng(d).size())
          ga_RegionError(ndim, lo, hi, *g_a);

      PatchTransfers i(ld, buf, dad, (void**) GA[handle].ptr, size, lo, hi) ;

//      i.loop(0, 1, 0, 0, 1, 1, dad->grp(), 0) ;  // enumerate block transfers.
      i.loop(0, 1, 0, 0, 1, 1, dad->grp().lead_prc(), 0) ;
                                          // enumerate block transfers.

      gaPermuteProcList(i.numBlk);

      int*   blocks  = i.getBlks() ;
      void** locPtrs = i.getLocs() ;
      void** remPtrs = i.getRems() ;

      for(idx=0; idx < i.numBlk; idx++) {
          p = (Integer)ProcListPerm [idx] ;
	  int*  block  = blocks + i.blockLen * p ;
	  void* locPtr = locPtrs [p] ;
	  void* remPtr = remPtrs [p] ;

          const int remId        = block [MBLKT_REM] ;
          const int numStrLevels = block [MBLKT_NST] ;

          int* exts    = block + MBLKT_CTG ;  // `contigLen' prepended
                                              // to vector.
          int* locStrs = exts + numStrLevels + 1 ;
          int* remStrs = locStrs + numStrLevels ;

          ARMCI_GetS(remPtr, remStrs, locPtr, locStrs, exts,
                     numStrLevels, remId) ;

          GAbytes.gettot += block [MBLKT_SIZ] ;
          if(remId == GAme)
            GAbytes.getloc += block [MBLKT_SIZ] ;
      }

      GA_POP_NAME;
}


void FATR  ga_get_(Integer* g_a, Integer* ilo, Integer* ihi,
                   Integer* jlo, Integer* jhi, Void* buf, Integer* ld)
{
Integer lo[2], hi[2];

#ifdef GA_TRACE
   trace_stime_();
#endif

   lo[0]=*ilo;
   lo[1]=*jlo;
   hi[0]=*ihi;
   hi[1]=*jhi;
   nga_get_(g_a, lo, hi, buf, ld);

#ifdef GA_TRACE
   trace_etime_();
   op_code = GA_OP_GET;
   trace_genrec_(g_a, ilo, ihi, jlo, jhi, &op_code);
#endif
}


/*\ local accumulate
\*/
void ga_acc_local(Integer g_a, Integer ilo, Integer ihi, Integer jlo, 
                  Integer jhi, void* buf, Integer offset, Integer ld_loc,
                  Integer proc, void* alpha)
{
char     *ptr_loc, *ptr_glob;
int      ld_glob, rows, cols, type = GA[GA_OFFSET + g_a].type;
int      size=GAsizeofM(type);
int      count[2], optype;

   GA_PUSH_NAME("ga_acc_local"); 

   gaShmemLocation(proc, g_a, ilo, jlo, &ptr_glob, &ld_glob);
   ptr_loc = (char *)buf   + size * offset;
   rows = ihi - ilo +1;
   cols = jhi - jlo +1;

   count[0] = rows*size; count[1]=cols;
   ld_loc *= size; ld_glob *= size;

   if(GA[GA_OFFSET + g_a].type==MT_F_DBL) optype= ARMCI_ACC_DBL;
   else if(type==MT_F_DCPL)optype= ARMCI_ACC_DCP;
   else if(size==sizeof(int))optype= ARMCI_ACC_INT;
   else ga_error("now is the time to add MT_F_LONG to armci!",0);

   int ld_loci = ld_loc ;
   ARMCI_AccS(optype, alpha, ptr_loc, &ld_loci, ptr_glob, &ld_glob, count, 1,proc);

   GA_POP_NAME;

}




/*\ ACCUMULATE OPERATION FOR A N-DIMENSIONAL PATCH OF GLOBAL ARRAY
 *
 *  g_a += alpha * patch
\*/
void FATR nga_acc_(Integer *g_a,
                   Integer *lo,
                   Integer *hi,
                   void    *buf,
                   Integer *ld,
                   void    *alpha)
{
Integer  p, handle=GA_OFFSET + *g_a;
Integer  idx, ndim, size, type ;
int optype;

      GA_PUSH_NAME("nga_acc");

      size = GA[handle].elemsize;
      type = GA[handle].type;

      if(type==MT_F_DBL) optype= ARMCI_ACC_DBL;
      else if(type==MT_F_DCPL)optype= ARMCI_ACC_DCP;
      else if(size==sizeof(int))optype= ARMCI_ACC_INT;
      else ga_error("now is the time to add MT_F_LONG to armci!",0);

      DAD* dad = &(GA[handle].dad) ;
      ndim = dad->rnk() ;

      GAstat.numacc++;

      for(int d = 0; d < ndim ; d++)
        if(lo [d] < 1 || hi[d] > dad->rng(d).size())
          ga_RegionError(ndim, lo, hi, *g_a);

      PatchTransfers i(ld, buf, dad, (void**) GA[handle].ptr, size, lo, hi) ;

//      i.loop(0, 1, 0, 0, 1, 1, dad->grp(), 0) ;  // enumerate block transfers.
      i.loop(0, 1, 0, 0, 1, 1, dad->grp().lead_prc(), 0) ;
                                       // enumerate block transfers.

      gaPermuteProcList(i.numBlk);

      int*   blocks  = i.getBlks() ;
      void** locPtrs = i.getLocs() ;
      void** remPtrs = i.getRems() ;

      for(idx=0; idx < i.numBlk; idx++) {
          p = (Integer)ProcListPerm [idx] ;
	  int*  block  = blocks + i.blockLen * p ;
	  void* locPtr = locPtrs [p] ;
	  void* remPtr = remPtrs [p] ;

          const int remId        = block [MBLKT_REM] ;
          const int numStrLevels = block [MBLKT_NST] ;

          int* exts    = block + MBLKT_CTG ;  // `contigLen' prepended
                                              // to vector.
          int* locStrs = exts + numStrLevels + 1 ;
          int* remStrs = locStrs + numStrLevels ;

          ARMCI_AccS(optype, alpha, locPtr, locStrs, remPtr, remStrs, exts,
                     numStrLevels, remId) ;

          GAbytes.acctot += block [MBLKT_SIZ] ;
          if(remId == GAme)
            GAbytes.accloc += block [MBLKT_SIZ] ;
      }

      GA_POP_NAME;
}



void FATR  ga_acc_(Integer* g_a, Integer* ilo, Integer* ihi,
                   Integer* jlo, Integer* jhi, void* buf, Integer* ld,
                   void* alpha)
{
Integer lo[2], hi[2];
#ifdef GA_TRACE
   trace_stime_();
#endif

   lo[0]=*ilo;
   lo[1]=*jlo;
   hi[0]=*ihi;
   hi[1]=*jhi;
   nga_acc_(g_a, lo, hi, buf, ld, alpha);

#ifdef GA_TRACE
   trace_etime_();
   op_code = GA_OP_ACC;
   trace_genrec_(g_a, ilo, ihi, jlo, jhi, &op_code);
#endif
}

void nga_access_ptr(Integer* g_a, Integer lo[], Integer hi[],
                      void* ptr, Integer ld[])

{
char *lptr;
Integer  handle = GA_OFFSET + *g_a;
Integer  ow ;

   GA_PUSH_NAME("nga_access_ptr");

   DAD* dad = &(GA[GA_OFFSET + *g_a].dad) ;
   const int rank = dad->rnk() ;

   for(int d = 0 ; d < rank ; d++) {
     Map   u = dad->map(d) ;
     Range x = u.rng() ;
     const int n = x.size() ;

     const int ihi = hi [d], ilo = lo [d] ;

     if(ilo > ihi)
       ga_RegionError(rank, lo, hi, *g_a);

     if(ihi < 1 || ihi > n)
       ga_error("locate top failed",0);

     const int x_crd = x.dim().crd() ;

     const int hi_crd = x(ihi - 1).crd ;  // (Could use a `ker_glb' member...)
     if(hi_crd != x_crd)
       ga_error("cannot access top of the patch",hi_crd) ;

     if(ilo < 1 || ilo > n)
       ga_error("locate bottom failed",0);

     const int lo_crd = x(ilo - 1).crd ;  // (Could use a `ker_glb' member...)
     if(lo_crd != x_crd)
       if(ow != GAme) ga_error("cannot access bottom of the patch",lo_crd);
   }

   gam_Location(ow,handle, lo, &lptr, ld);
   *(char**)ptr = lptr;
   GA_POP_NAME;
}


/*\ PROVIDE ACCESS TO A PATCH OF A GLOBAL ARRAY
\*/
void FATR nga_access_(Integer* g_a, Integer lo[], Integer hi[],
                      Integer* index, Integer ld[])
{
char *ptr;
Integer  handle = GA_OFFSET + *g_a;
Integer  ow ;

   GA_PUSH_NAME("nga_access");

   DAD* dad = &(GA[GA_OFFSET + *g_a].dad) ;
   const int rank = dad->rnk() ;

   int offset = 0 ;
   for(int d = 0 ; d < rank ; d++) {
     Map   u = dad->map(d) ;
     Range x = u.rng() ;
     const int n = x.size() ;

     const int ihi = hi [d], ilo = lo [d] ;

     if(ilo > ihi)
       ga_RegionError(rank, lo, hi, *g_a);

     if(ihi < 1 || ihi > n)
       ga_error("locate top failed",0);

     const int x_crd = x.dim().crd() ;

     const int hi_crd = x(ihi - 1).crd ;  // (Could use a `ker_glb' member...)
     if(hi_crd != x_crd) {
       ga_error("cannot access top of the patch",hi_crd) ;
     }

     if(ilo < 1 || ilo > n)
       ga_error("locate bottom failed",0);

     Location i = x(ilo - 1) ;
     if(i.crd != x_crd)
       if(ow != GAme) ga_error("cannot access bottom of the patch",i.crd);

//AdlibPrintf("nga_access_: d = %d offset = %d\n", d, u.offset(i)) ;  // debug
     offset += u.offset(i) ;

     ld [d] = x.volume() ;
   }

   ptr = GA[handle].ptr[GAme] + GA[GA_OFFSET + *g_a].elemsize * offset ;

   /*
    * return patch address as the distance in bytes from the reference address
    *
    * .in Fortran we need only the index to the type array: dbl_mb or int_mb
    *  that are elements of COMMON in the the mafdecls.h include file
    * .in C we need both the index and the pointer
    */

   /* compute index and check if it is correct */
   switch (GA[handle].type){
     case MT_F_DBL:
        *index = (Integer) (ptr - (char*)DBL_MB);
        if(ptr != ((char*)DBL_MB)+ *index ){
               ga_error("ga_access: MA addressing problem dbl - index",handle);
        }
        break;

     case MT_F_DCPL:
        *index = (Integer) (ptr - (char*)DCPL_MB);
        if(ptr != ((char*)DCPL_MB)+ *index ){
              ga_error("ga_access: MA addressing problem dcpl - index",handle);
        }
        break;

     case MT_F_INT:
        *index = (Integer) (ptr - (char*)INT_MB);
        if(ptr != ((char*)INT_MB) + *index) {
               ga_error("ga_access: MA addressing problem int - index",handle);
        }
        break;
   }

   /* check the allignment */
   if(*index % GA[handle].elemsize){
       fprintf(stderr,"index=%ld size=%ld off =%ld\n",
              (long)*index, (long)GA[handle].elemsize,(long)*index%GA[handle].elemsize);
       ga_error(" ga_access: base address misallignment ",(long)index);
   }

   /* adjust index according to the data type */
   *index /= GA[handle].elemsize;

   /* adjust index for Fortran addressing */
   (*index) ++ ;
   FLUSH_CACHE;

   GA_POP_NAME;
}


/*\ PROVIDE ACCESS TO A PATCH OF A GLOBAL ARRAY
\*/
void FATR  ga_access_(Integer* g_a, Integer* ilo, Integer* ihi,
                      Integer* jlo, Integer* jhi, Integer* index, Integer* ld)
{
Integer lo[2], hi[2];
     lo[0]=*ilo;
     lo[1]=*jlo;
     hi[0]=*ihi;
     hi[1]=*jhi;
     nga_access_(g_a,lo,hi,index,ld);
}


/*\ RELEASE ACCESS TO A PATCH OF A GLOBAL ARRAY
\*/
void FATR  ga_release_(Integer* g_a, Integer* ilo, Integer* ihi,
                       Integer* jlo, Integer* jhi)
{}

/*\ RELEASE ACCESS TO A PATCH OF A GLOBAL ARRAY
\*/
void FATR  nga_release_(Integer* g_a, Integer* lo, Integer* hi)
{}


/*\ RELEASE ACCESS & UPDATE A PATCH OF A GLOBAL ARRAY
\*/
void FATR  ga_release_update_(Integer* g_a, Integer* ilo, Integer* ihi,
                              Integer* jlo, Integer* jhi)
{}


/*\ RELEASE ACCESS & UPDATE A PATCH OF A GLOBAL ARRAY
\*/
void FATR  nga_release_update_(Integer* g_a, Integer* lo, Integer* hi)
{}



/*\ INQUIRE POPERTIES OF A GLOBAL ARRAY
 *  Fortran version
\*/ 
void FATR  ga_inquire_(Integer* g_a, Integer* type,
                       Integer* dim1, Integer* dim2)
{
   ga_check_handleM(g_a, "ga_inquire");
   *type       = GA[GA_OFFSET + *g_a].type;
   DAD* dad = &(GA[GA_OFFSET + *g_a].dad) ;
   *dim1       = dad->rng(0).size() ;
   *dim2       = dad->rng(1).size() ;
}

void FATR nga_inquire_(Integer *g_a, Integer *type, Integer *ndim,Integer *dims)
{
Integer handle = GA_OFFSET + *g_a,i;
   ga_check_handleM(g_a, "nga_inquire");
   *type       = GA[handle].type;
   DAD* dad = &(GA[GA_OFFSET + *g_a].dad) ;
   *ndim       = dad->rnk() ;
   for(i=0;i<*ndim;i++)dims[i]=dad->rng(i).size() ;
}



/*\ INQUIRE NAME OF A GLOBAL ARRAY
 *  Fortran version
\*/
#if defined(CRAY) || defined(WIN32)
void FATR  ga_inquire_name_(Integer* g_a, _fcd array_name)
{
   c2fstring(GA[GA_OFFSET+ *g_a].name,_fcdtocp(array_name),_fcdlen(array_name));
}
#else
void FATR  ga_inquire_name_(Integer* g_a, char* array_name, int len)
{
   c2fstring(GA[GA_OFFSET + *g_a].name, array_name, len);
}
#endif




/*\ INQUIRE NAME OF A GLOBAL ARRAY
 *  C version
\*/
void ga_inquire_name(Integer* g_a, char** array_name)
{ 
   ga_check_handleM(g_a, "ga_inquire_name");
   *array_name = GA[GA_OFFSET + *g_a].name;
}





/*\ RETURN COORDINATES OF ARRAY BLOCK HELD BY A PROCESSOR
 *
 * In general Adlib allows more complex strategies for mapping logical
 * process coordinates to physical processors than GA, and the reverse
 * mapping -- physical to logical -- isn't a function that can be computed
 * efficiently with the documented Adlib interface.  This implementation 
 * relies on the undocumented fact that, *by default*, Adlib maps process
 * coordinates to physical processors in the same way as GA.
\*/
void FATR nga_proc_topology_(Integer* g_a, Integer* proc, Integer* subscript)
{
Integer d, index, ndim, ga_handle = GA_OFFSET + *g_a ;

   ga_check_handleM(g_a, "nga_proc_topology");

   DAD* dad = &(GA[ga_handle].dad) ;
   const Procs* p = dad->grp().prc() ;

   ndim = dad->rnk() ;

   for(d=0, index = *proc; d<ndim; d++){
       int nblock = p->dim(d).size() ;
       subscript[d] = index % nblock ;
       index  /= nblock ;  
   }
}


/*\ LOCATE THE OWNER OF SPECIFIED ELEMENT OF A GLOBAL ARRAY
\*/
logical FATR nga_locate_(Integer *g_a, Integer* subscript, Integer* owner)
{
Integer d, ndim, ga_handle = GA_OFFSET + *g_a ;

   ga_check_handleM(g_a, "nga_locate");
   DAD* dad = &(GA[ga_handle].dad) ;
   ndim = dad->rnk() ;

   Group p = dad->grp() ;
   for(d = 0, *owner = -1; d < ndim; d++) {
       Range x = dad->rng(d) ;
       if(subscript [d] < 1 || subscript [d] > x.size()) return FALSE;

       //p /= x(subscript [d] - 1) ;  // Doesn't work, for unknown reasons.

       Location i = x(subscript [d] - 1) ;
       p.restrict(i.dim, i.crd) ;
   }
   *owner = p.lead_abs() ;

   return TRUE;
}


/*\ LOCATE PROCESSORS/OWNERS OF THE SPECIFIED PATCH OF A GLOBAL ARRAY
\*/
logical FATR nga_locate_region_( Integer *g_a,
                                 Integer *lo,
                                 Integer *hi,
                                 Integer *map,
                                 Integer *proclist,
                                 Integer *np)
{
Integer  ga_handle;
Integer  d ;

   ga_check_handleM(g_a, "nga_locate_region");

   ga_handle = GA_OFFSET + *g_a;

   DAD* dad = &(GA[ga_handle].dad) ;

   for(d = 0; d < dad->rnk() ; d++)
       if(lo[d] < 1 || hi[d]> dad->rng(d).size()) return FALSE;

   PatchOwners i(dad, lo, hi, map, proclist) ;
   *np = 0 ;
   i.loop(dad->grp(), 0, np) ;

   return(TRUE);
}
    

void ga_scatter_acc_local(Integer g_a, Void *v,Integer *i,Integer *j,
                          Integer nv, void* alpha, Integer proc) 
{
void **ptr_src, **ptr_dst;
char *ptr_ref;
Integer ldp, item_size, ilo, ihi, jlo, jhi, type;
armci_giov_t desc;
register Integer k, offset;
int rc;

  if (nv < 1) return;

  GA_PUSH_NAME("ga_scatter_local");

  ga_distribution_(&g_a, &proc, &ilo, &ihi, &jlo, &jhi);

  /* get address of the first element owned by proc */
  gaShmemLocation(proc, g_a, ilo, jlo, &ptr_ref, &ldp);

  type = GA[GA_OFFSET + g_a].type;
  item_size = GAsizeofM(type);

  ptr_src = (void**) gai_malloc(2*nv*sizeof(void*));
  if(ptr_src==NULL)ga_error("gai_malloc failed",nv);
  else ptr_dst=ptr_src+ nv;

  for(k=0; k< nv; k++){
     if(i[k] < ilo || i[k] > ihi  || j[k] < jlo || j[k] > jhi){
       sprintf(err_string,"proc=%d invalid i/j=(%d,%d)>< [%d:%d,%d:%d]",
               proc, i[k], j[k], ilo, ihi, jlo, jhi); 
       ga_error(err_string,g_a);
     }

     offset  = (j[k] - jlo)* ldp + i[k] - ilo;
     ptr_dst[k] = ptr_ref + item_size * offset;
     ptr_src[k] = ((char*)v) + k*item_size;
  }
  desc.bytes = item_size;
  desc.src_ptr_array = ptr_src;
  desc.dst_ptr_array = ptr_dst;
  desc.ptr_array_len = nv;

  if(alpha ==NULL) rc=ARMCI_PutV(&desc, 1, (int)proc);
  else{

    int optype;
    if(type==MT_F_DBL) optype= ARMCI_ACC_DBL;
    else if(type==MT_F_DCPL)optype= ARMCI_ACC_DCP;
    else if(item_size==sizeof(int))optype= ARMCI_ACC_INT;
    else ga_error("now is the time to add MT_F_LONG to ARMCI!",0);
    rc= ARMCI_AccV(optype, alpha, &desc, 1, (int)proc);
  }

  if(rc) ga_error("scatter/_acc failed in armci",rc);

  gai_free(ptr_src);

  GA_POP_NAME;
}




/*\ SCATTER OPERATION elements of v into the global array
\*/
void FATR  ga_scatter_(Integer* g_a, Void* v,
                       Integer* i, Integer* j, Integer* nv)
{
register Integer k;
Integer pindex, phandle, item_size;
Integer first, nelem, proc, type=GA[GA_OFFSET + *g_a].type;

  if (*nv < 1) return;

  ga_check_handleM(g_a, "ga_scatter");
  GA_PUSH_NAME("ga_scatter");
  GAstat.numsca++;

  if(!MA_push_get(MT_F_INT,*nv, "ga_scatter--p", &phandle, &pindex))
            ga_error("MA alloc failed ", *g_a);

  /* find proc that owns the (i,j) element; store it in temp: INT_MB[] */
  for(k=0; k< *nv; k++) if(! ga_locate_(g_a, i+k, j+k, INT_MB+pindex+k)){
         sprintf(err_string,"invalid i/j=(%d,%d)", i[k], j[k]);
         ga_error(err_string,*g_a);
  }

  /* determine limit for message size --  v,i, & j will travel together */
  item_size = GAsizeofM(type);
  GAbytes.scatot += (double)item_size**nv ;

  /* Sort the entries by processor */
  ga_sort_scat(nv, (DoublePrecision*)v, i, j, INT_MB+pindex, type );

  /* go through the list again executing scatter for each processor */

  first = 0;
  do { 
      proc  = INT_MB[pindex+first]; 
      nelem = 0;

      /* count entries for proc from "first" to last */
      for(k=first; k< *nv; k++){
        if(proc == INT_MB[pindex+k]) nelem++;
        else break;
      }

      if(proc == GAme){
             GAbytes.scaloc += (double)item_size* nelem ;
      }

      ga_scatter_acc_local(*g_a, ((char*)v)+item_size*first, i+first, 
                        j+first, nelem, NULL, proc);
      first += nelem;

  }while (first< *nv);

  if(! MA_pop_stack(phandle)) ga_error(" pop stack failed!",phandle);

  GA_POP_NAME;
}
      


/*\ SCATTER OPERATION elements of v into the global array
\*/
void FATR  ga_scatter_acc_(Integer* g_a, Void* v, Integer* i, Integer* j,
                           Integer* nv, Void* alpha)
{
register Integer k;
Integer pindex, phandle, item_size;
Integer first, nelem, proc, type=GA[GA_OFFSET + *g_a].type;

  if (*nv < 1) return;

  ga_check_handleM(g_a, "ga_scatter_acc");
  GA_PUSH_NAME("ga_scatter_acc");
  GAstat.numsca++;

  if(!MA_push_get(MT_F_INT,*nv, "ga_scatter_acc--p", &phandle, &pindex))
            ga_error("MA alloc failed ", *g_a);

  /* find proc that owns the (i,j) element; store it in temp: INT_MB[] */
  for(k=0; k< *nv; k++) if(! ga_locate_(g_a, i+k, j+k, INT_MB+pindex+k)){
         sprintf(err_string,"invalid i/j=(%d,%d)", i[k], j[k]);
         ga_error(err_string,*g_a);
  }

  /* determine limit for message size --  v,i, & j will travel together */
  item_size = GAsizeofM(type);
  GAbytes.scatot += (double)item_size**nv ;

  /* Sort the entries by processor */
  ga_sort_scat(nv, (DoublePrecision*)v, i, j, INT_MB+pindex, type );

  /* go through the list again executing scatter for each processor */

  first = 0;
  do {
      proc  = INT_MB[pindex+first];
      nelem = 0;

      /* count entries for proc from "first" to last */
      for(k=first; k< *nv; k++){
        if(proc == INT_MB[pindex+k]) nelem++;
        else break;
      }

      if(proc == GAme){
             GAbytes.scaloc += (double)item_size* nelem ;
      }

      ga_scatter_acc_local(*g_a, ((char*)v)+item_size*first, i+first,
                           j+first, nelem, alpha, proc);
      first += nelem;

  }while (first< *nv);

  if(! MA_pop_stack(phandle)) ga_error(" pop stack failed!",phandle);

  GA_POP_NAME;
}


extern "C" {
  extern void ga_sort_permutation(Integer* pn, Integer* index, Integer* base) ;
}

/*\ permutes input index list using sort routine used in scatter/gather
\*/
void FATR  ga_sort_permut_(Integer* g_a, Integer* index,
                           Integer* i, Integer* j, Integer* nv)
{
register Integer k;
Integer pindex, phandle;

  if (*nv < 1) return;

  if(!MA_push_get(MT_F_INT,*nv, "ga_sort_permut--p", &phandle, &pindex))
            ga_error("MA alloc failed ", *g_a);

  /* find proc that owns the (i,j) element; store it in temp: INT_MB[] */
  for(k=0; k< *nv; k++) if(! ga_locate_(g_a, i+k, j+k, INT_MB+pindex+k)){
         sprintf(err_string,"invalid i/j=(%d,%d)", i[k], j[k]);
         ga_error(err_string,*g_a);
  }

  /* Sort the entries by processor */
  ga_sort_permutation(nv, index, INT_MB+pindex);
  if(! MA_pop_stack(phandle)) ga_error(" pop stack failed!",phandle);
}



void ga_gather_local(Integer g_a, Void* v, Integer* i, Integer* j,
                     Integer nv, Integer proc) 
{
void **ptr_src, **ptr_dst;
char *ptr_ref;
Integer ldp, item_size, ilo, ihi, jlo, jhi;
armci_giov_t desc;
register Integer k, offset;

  if (nv < 1) return;

  GA_PUSH_NAME("ga_gather_local");

  ga_distribution_(&g_a, &proc, &ilo, &ihi, &jlo, &jhi);

  /* get address of the first element owned by proc */
  gaShmemLocation(proc, g_a, ilo, jlo, &ptr_ref, &ldp);

  item_size = GAsizeofM(GA[GA_OFFSET + g_a].type);
  ptr_src = (void**) gai_malloc(2*nv*sizeof(void*));
  if(ptr_src==NULL)ga_error("gai_malloc failed",nv);
  else ptr_dst=ptr_src+ nv;

  for(k=0; k< nv; k++){
     if(i[k] < ilo || i[k] > ihi  || j[k] < jlo || j[k] > jhi){
       sprintf(err_string,"proc=%d invalid i/j=(%d,%d)>< [%d:%d,%d:%d]",
               proc, i[k], j[k], ilo, ihi, jlo, jhi); 
       ga_error(err_string,g_a);
     }

     offset  = (j[k] - jlo)* ldp + i[k] - ilo;
     ptr_dst[k] = ((char*)v) + k*item_size;
     ptr_src[k] = ptr_ref + item_size * offset;
  }
  desc.bytes = item_size;
  desc.src_ptr_array = ptr_src;
  desc.dst_ptr_array = ptr_dst;
  desc.ptr_array_len = nv;

  ARMCI_GetV(&desc, 1, (int)proc);

  gai_free(ptr_src);

  GA_POP_NAME;
}


/*\ GATHER OPERATION elements from the global array into v
\*/
void FATR  ga_gather_(Integer* g_a, Void* v, Integer* i, Integer* j,
                      Integer* nv)
{
register Integer k, nelem;
Integer pindex, phandle, item_size;
Integer first, proc;

  if (*nv < 1) return;

  ga_check_handleM(g_a, "ga_gather");
  GA_PUSH_NAME("ga_gather");
  GAstat.numgat++;


  if(!MA_push_get(MT_F_INT, *nv, "ga_gather--p", &phandle, &pindex))
            ga_error("MA failed ", *g_a);

  /* find proc that owns the (i,j) element; store it in temp: INT_MB[] */
  for(k=0; k< *nv; k++) if(! ga_locate_(g_a, i+k, j+k, INT_MB+pindex+k)){
       sprintf(err_string,"invalid i/j=(%d,%d)", i[k], j[k]);
       ga_error(err_string,*g_a);
  }

  /* Sort the entries by processor */
  ga_sort_gath_(nv, i, j, INT_MB+pindex);
   
  item_size = GA[GA_OFFSET + *g_a].elemsize;
  GAbytes.gattot += (double)item_size**nv;


  /* go through the list again executing gather for each processor */

  first = 0;
  do {
      proc  = INT_MB[pindex+first];
      nelem = 0;

      /* count entries for proc from "first" to last */
      for(k=first; k< *nv; k++){
        if(proc == INT_MB[pindex+k]) nelem++;
        else break;
      }

      if(proc == GAme){
             GAbytes.gatloc += (double)item_size* nelem ;
      }

      ga_gather_local(*g_a, ((char*)v)+item_size*first, i+first, j+first,
                       nelem, proc);
      first += nelem; 

  }while (first< *nv);

  if(! MA_pop_stack(phandle)) ga_error(" pop stack failed!",phandle);

  GA_POP_NAME;
}
      
           



/*\ READ AND INCREMENT AN ELEMENT OF A GLOBAL ARRAY
\*/
Integer FATR nga_read_inc_(Integer* g_a, Integer* subscript, Integer* inc)
{
Integer *ptr, ldp[MAXDIM], value, proc, handle=GA_OFFSET+*g_a;
int optype;

    ga_check_handleM(g_a, "nga_read_inc");
    GA_PUSH_NAME("ga_read_inc");

    if(GA[handle].type!=MT_F_INT) ga_error("type must be integer",*g_a);

    GAstat.numrdi++;
    GAbytes.rditot += (double)sizeof(Integer);

    /* find out who owns it */
    nga_locate_(g_a, subscript, &proc);

    /* get an address of the g_a(subscript) element */
/*    gaShmemLocation(proc, *g_a, subscript[0],subscript[1],(char**)&ptr,&ldp[0]);*/
    gam_Location(proc, handle,  subscript, (char**)&ptr, ldp);

#   ifdef EXT_INT
      optype = ARMCI_FETCH_AND_ADD_LONG;
#   else
      optype = ARMCI_FETCH_AND_ADD;
#   endif

    ARMCI_Rmw(optype, (int*)&value, (int*)ptr, (int)*inc, (int)proc);

    if(GAme == proc)GAbytes.rdiloc += (double)sizeof(Integer);

   GA_POP_NAME;
   return(value);
}



Integer FATR ga_nodeid_()
{
  return ((Integer)GAme);
}


Integer FATR ga_nnodes_()
{
  return ((Integer)GAnproc);
}



/*\ COMPARE DISTRIBUTIONS of two global arrays
\*/
logical FATR ga_compare_distr_(Integer *g_a, Integer *g_b)
{
int h_a =*g_a + GA_OFFSET;
int h_b =*g_b + GA_OFFSET;

   GA_PUSH_NAME("ga_compare_distr");
   ga_check_handleM(g_a, "distribution a");
   ga_check_handleM(g_b, "distribution b");

   GA_POP_NAME;

   DAD* a = &(GA[h_a].dad) ;
   DAD* b = &(GA[h_b].dad) ;

   if(!a->grp().isEqual(b->grp())) return FALSE ;
   if(a->rnk() != b->rnk()) return FALSE ;
   for(int d = 0 ; d < a->rnk() ; d++)
     if(!a->rng(d).isAligned(b->rng(d))) return FALSE ;

   return TRUE;
}


/*************************************************************************/

logical FATR ga_locate_region_(Integer* g_a,
                               Integer* ilo, Integer* ihi,
                               Integer* jlo, Integer* jhi,
                               Integer mapl [] [5], Integer* np)
{
   logical status;
   Integer lo[2], hi[2], p;
   lo[0]=*ilo; lo[1]=*jlo;
   hi[0]=*ihi; hi[1]=*jhi;

   status = nga_locate_region_(g_a,lo,hi, (Integer*) map, proclist, np);

   /* need to swap elements (ilo,jlo,ihi,jhi) -> (ilo,ihi,jlo,jhi) */
   for(p = 0; p< *np; p++){
     mapl[p][0] = map[4*p];
     mapl[p][1] = map[4*p + 2];
     mapl[p][2] = map[4*p + 1];
     mapl[p][3] = map[4*p + 3];
     mapl[p][4] = proclist[p];
   }

   return status;
}



void ga_scatter_local(Integer g_a, Void *v,Integer *i,Integer *j,
                      Integer nv, Integer proc) 
{
     ga_error("ga_scatter_local should not be used",0);
}


/*\ LOCATE THE OWNER OF THE (i,j) ELEMENT OF A GLOBAL ARRAY
\*/
logical FATR ga_locate_(Integer* g_a, Integer* i, Integer* j, Integer* owner)
{
Integer subscript[2];
  
    subscript[0]=*i; subscript[1]=*j;

    return nga_locate_(g_a, subscript, owner);
}


/*\ RETURN COORDINATES OF A 2-D GA PATCH ASSOCIATED WITH PROCESSOR proc
\*/
void FATR  ga_distribution_(Integer* g_a, Integer* proc,
                            Integer* ilo, Integer* ihi,
                            Integer* jlo, Integer* jhi)
{
Integer  lo[2], hi[2];

   nga_distribution_(g_a, proc, lo, hi);
   *ilo = lo[0]; *ihi=hi[0];
   *jlo = lo[1]; *jhi=hi[1]; 
   
}


/*\ RETURN COORDINATES OF ARRAY BLOCK HELD BY A PROCESSOR
\*/
void FATR ga_proc_topology_(Integer* g_a, Integer* proc,
                            Integer* pr, Integer* pc)
{
Integer subscript[2];
   nga_proc_topology_(g_a, proc,subscript);
   *pr = subscript[0]; 
   *pc = subscript[1]; 
}



Integer ga_read_inc_local(Integer g_a, Integer i, Integer j,
                          Integer inc, Integer proc)
{
   ga_error("ga_rdi_local should not be used",0);
   return(-1);
}


/*\ READ AND INCREMENT AN ELEMENT OF A GLOBAL ARRAY
\*/
Integer FATR ga_read_inc_(Integer* g_a, Integer* i, Integer* j, Integer* inc)
{
Integer  value, subscript[2];

#ifdef GA_TRACE
       trace_stime_();
#endif

   subscript[0] =*i;
   subscript[1] =*j;
   value = nga_read_inc_(g_a, subscript, inc);

#  ifdef GA_TRACE
     trace_etime_();
     op_code = GA_OP_RDI;
     trace_genrec_(g_a, i, i, j, j, &op_code);
#  endif

   return(value);
}


/*********************** other utility routines *************************/

extern "C" {
  void ga_ma_get_ptr_(char** ptr, char* address) ;
  Integer ga_ma_diff_(char* ptr1, char* ptr2) ;
}

void ga_ma_get_ptr_(char** ptr, char* address)
{
   *ptr = address;
}


Integer ga_ma_diff_(char* ptr1, char* ptr2)
{
   return((Integer)(ptr2-ptr1));
}


/*************************************************************************/

// Things to do
//
//   `gam_Location' is fully DAD-compatible.

