/* 
 * module: global.armci.c
 * author: Jarek Nieplocha
 * description: implements GA primitive operations --
 *              create (regular& irregular) and duplicate, destroy, get, put,
 *              accumulate, scatter, gather, read&increment & synchronization 
 * 
 * DISCLAIMER
 *
 * This material was prepared as an account of work sponsored by an
 * agency of the United States Government.  Neither the United States
 * Government nor the United States Department of Energy, nor Battelle,
 * nor any of their employees, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
 * ASSUMES ANY LEGAL LIABILITY OR RESPONSIBILITY FOR THE ACCURACY,
 * COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
 * SOFTWARE, OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT
 * INFRINGE PRIVATELY OWNED RIGHTS.
 *
 *
 * ACKNOWLEDGMENT
 *
 * This software and its documentation were produced with United States
 * Government support under Contract Number DE-AC06-76RLO-1830 awarded by
 * the United States Department of Energy.  The United States Government
 * retains a paid-up non-exclusive, irrevocable worldwide license to
 * reproduce, prepare derivative works, perform publicly and display
 * publicly by or for the US Government, including the right to
 * distribute to other US Government contractors.
 */

 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "global.h"
#include "globalp.h"
#include "message.h"
#include "macdecls.h"
#include "global.armci2.h"

#define DEBUG 0
#define USE_MALLOC 1
#define INVALID_MA_HANDLE -1 
#define NEAR_INT(x) (x)< 0.0 ? ceil( (x) - 0.5) : floor((x) + 0.5)

/*uncomment line below to initialize arrays in ga_create/duplicate */
/*#define GA_CREATE_INDEF yes */

/*uncomment line below to verify consistency of MA in every sync */
/*#define CHECK_MA yes */

/* uncomment line below to verify if MA base address is alligned wrt datatype*/ 
#if !(defined(LINUX) || defined(CRAY))
#define CHECK_MA_ALGN 1
#endif



/*\ SYNCHRONIZE ALL THE PROCESSES
\*/
void FATR ga_sync_()
{
extern int GA_fence_set;
#ifdef CHECK_MA
Integer status;
#endif

       GA_fence_set=0;
       ARMCI_AllFence();
       ga_msg_sync_();
#ifdef CHECK_MA
       status = MA_verify_allocator_stuff();
#endif
}


Integer GAsizeof(type)    
        Integer type;
{
  switch (type) {
     case MT_F_DBL  : return (sizeof(DoublePrecision));
     case MT_F_INT  : return (sizeof(Integer));
     case MT_F_DCPL : return (sizeof(DoubleComplex));
          default   : return 0; 
  }
}


/*\ FINAL CLEANUP of shmem when terminating
\*/
void ga_clean_resources()
{
    ARMCI_Cleanup();
}


/*\ CHECK GA HANDLE and if it's wrong TERMINATE
 *  Fortran version
\*/
#if defined(CRAY) || defined(WIN32)
void FATR  ga_check_handle_(g_a, fstring)
     Integer *g_a;
     _fcd fstring;
#else
void FATR  ga_check_handle_(g_a, fstring,slen)
     Integer *g_a;
     int  slen;
     char *fstring;
#endif
{
char  buf[FLEN];

    if( GA_OFFSET + (*g_a) < 0 || GA_OFFSET + (*g_a) >= max_global_array){
#if defined(CRAY) || defined(WIN32)
      f2cstring(_fcdtocp(fstring), _fcdlen(fstring), buf, FLEN);
#else
      f2cstring(fstring, slen, buf, FLEN);
#endif
      fprintf(stderr, " ga_check_handle: %s ", buf);
      ga_error(" invalid global array handle ", (*g_a));
    }
    if( ! (GA[GA_OFFSET + (*g_a)].actv) ){
#if defined(CRAY) || defined(WIN32)
      f2cstring(_fcdtocp(fstring), _fcdlen(fstring), buf, FLEN);
#else
      f2cstring(fstring, slen, buf, FLEN);
#endif
      ga_error(" global array is not active ", (*g_a));
    }
}



/*\ CHECK GA HANDLE and if it's wrong TERMINATE
 *  C version
\*/
void ga_check_handle(Integer *g_a,char * string)
{
  ga_check_handleM(g_a,string);
}




/*\ prepare permuted list of processes for remote ops
\*/
#define gaPermuteProcList(nproc)\
{\
  if((nproc) ==1) ProcListPerm[0]=0; \
  else{\
    int _i, iswap, temp;\
    if((nproc) > GAnproc) ga_error("permute_proc: error ", (nproc));\
\
    /* every process generates different random sequence */\
    (void)srand((unsigned)GAme); \
\
    /* initialize list */\
    for(_i=0; _i< (nproc); _i++) ProcListPerm[_i]=_i;\
\
    /* list permutation generated by random swapping */\
    for(_i=0; _i< (nproc); _i++){ \
      iswap = (int)(rand() % (nproc));  \
      temp = ProcListPerm[iswap]; \
      ProcListPerm[iswap] = ProcListPerm[_i]; \
      ProcListPerm[_i] = temp; \
    } \
  }\
}
     


/*\ INITIALIZE GLOBAL ARRAY STRUCTURES
 *
 *  either ga_initialize_ltd or ga_initialize must be the first 
 *         GA routine called (except ga_uses_ma)
\*/
void FATR  ga_initialize_()
{
Integer  i;
#ifdef CHECK_MA_ALGN
Integer  off_dbl, off_int, off_dcpl;
#endif

    if(GAinitialized) return;

    /* zero in pointers in GA array */
    for(i=0;i<MAX_ARRAYS; i++) {
       GA[i].ptr  = (char**)0;
       GA[i].mapc = (int*)0;
    }
    GAmaster= 0;
    GAnproc = (Integer)ga_msg_nnodes_();
    GAme = (Integer)ga_msg_nodeid_();
    MPme= ga_msg_nodeid_();
    MPnproc = ga_msg_nnodes_();

    if(GAnproc > MAX_NPROC && MPme==0){
      fprintf(stderr,"Current GA setup is for up to %d processors\n",MAX_NPROC);
      fprintf(stderr,"Please change MAX_NPROC in globalp.h & recompile\n");
      ga_error("terminating...",0);
    }


    /* set activity status for all arrays to inactive */
    for(i=0;i<max_global_array;i++)GA[i].actv=0;

    GAinitialized = 1;

    /* Initialize MA-like addressing:
     *    get addressees for the base arrays for double, complex and int types
     *
     * MA include files: macommon.h, macdecls.h and mafdecls.h
     */
     INT_MB = (Integer*)MA_get_mbase(MT_F_INT);
     DBL_MB = (DoublePrecision*)MA_get_mbase(MT_F_DBL);
     DCPL_MB= (DoubleComplex*)MA_get_mbase(MT_F_DCPL);

#   ifdef CHECK_MA_ALGN
        off_dbl = 0 != ((long)DBL_MB)%sizeof(DoublePrecision);
        off_int = 0 != ((long)INT_MB)%sizeof(Integer);
        off_dcpl= 0 != ((long)DCPL_MB)%sizeof(DoublePrecision);

        if(off_dbl)
           ga_error("ga_initialize: MA DBL_MB not alligned", (Integer)DBL_MB);

        if(off_int)
           ga_error("ga_initialize: INT_MB not alligned", (Integer)INT_MB);

        if(off_dcpl)
          ga_error("ga_initialize: DCPL_MB not alligned", (Integer)DCPL_MB);
#   endif

    if(DEBUG)
        printf("%d INT_MB=%d(%x) DBL_MB=%ld(%lx) DCPL_MB=%d(%lx)\n",
                GAme, INT_MB,INT_MB, DBL_MB,DBL_MB, DCPL_MB,DCPL_MB);
    ARMCI_Init();

}




/*\ IS MA USED FOR ALLOCATION OF GA MEMORY ?
\*/ 
logical FATR ga_uses_ma_()
{
#  if defined(CRAY_T3E) || defined(SP) || defined(LAPI) || defined(PARAGON)
     return FALSE;
#  else
     return TRUE;
#  endif
}


/*\ IS MEMORY LIMIT SET ?
\*/
logical FATR ga_memory_limited_()
{
   if(GA_memory_limited) return TRUE;
   else                  return FALSE;
}



/*\ RETURNS AMOUNT OF MEMORY on each processor IN ACTIVE GLOBAL ARRAYS 
\*/
Integer  FATR ga_inquire_memory_()
{
Integer i, sum=0;
    for(i=0; i<max_global_array; i++) 
        if(GA[i].actv) sum += GA[i].size; 
    return(sum);
}


/*\ RETURNS AMOUNT OF GA MEMORY AVAILABLE on calling processor 
\*/
Integer FATR ga_memory_avail_()
{
#ifdef SYSV
   return(GA_total_memory); 
#else
   Integer ma_limit = MA_inquire_avail(MT_F_BYTE);

   if ( GA_memory_limited ) return( MIN(GA_total_memory, ma_limit) );
   else return( ma_limit );
#endif
}


#define MBUFLEN 256
#define MBUF_LEN MBUFLEN+2
static double ga_int_malloc_buf[MBUF_LEN];
static int mbuf_used=0;
#define MBUF_GUARD -1998.1998
void *gai_malloc(int bytes)
{
    void *ptr;
    if(!mbuf_used && bytes <= MBUF_LEN){
       if(DEBUG){
          ga_int_malloc_buf[0]= MBUF_GUARD;
          ga_int_malloc_buf[MBUFLEN]= MBUF_GUARD;
       }
       ptr = ga_int_malloc_buf+1;
       mbuf_used++;
    }else{
        Integer handle, idx, elems = (bytes+sizeof(double)-1)/sizeof(double)+1; 
        if(MA_push_get(MT_DBL, elems, "temp", &handle, &idx)){
            MA_get_pointer(handle, &ptr);
            *((Integer*)ptr)= handle;
            ptr = ((double*)ptr)+ 1;  /* will break if sizeof(double)<sizeof(Integer) */
        }else
            ptr=NULL;
    }
    return ptr;
}

void gai_free(void *ptr)
{
    if(ptr == (ga_int_malloc_buf+1)){
        if(DEBUG){
          assert(ga_int_malloc_buf[0]== MBUF_GUARD);
          assert(ga_int_malloc_buf[MBUFLEN]== MBUF_GUARD);
          assert(mbuf_used ==1);
        }
        mbuf_used =0;
    }else{
        Integer handle= *( (Integer*) (-1 + (double*)ptr));
        if(!MA_pop_stack(handle)) ga_error("gai_free:MA_pop_stack failed",0);
    }
}

        

/*\ INITIALIZE GLOBAL ARRAY STRUCTURES and SET LIMIT ON GA MEMORY USAGE
 *  
 *  the byte limit is per processor (even for shared memory)
 *  either ga_initialize_ltd or ga_initialize must be the first 
 *         GA routine called (except ga_uses_ma)
 *  ga_initialize_ltd is another version of ga_initialize 
 *         without memory control
 *  mem_limit < 0 means "memory unlimited"
\*/
void FATR  ga_initialize_ltd_(mem_limit)
Integer *mem_limit;
{

  GA_total_memory = *mem_limit; 
  if(*mem_limit >= 0) GA_memory_limited = 1; 
  ga_initialize_();
}




/*\ CREATE A GLOBAL ARRAY
\*/
logical ga_create(type, dim1, dim2, array_name, chunk1, chunk2, g_a)
     Integer *type, *dim1, *dim2, *chunk1, *chunk2, *g_a; 
     char *array_name;
     /*
      * array_name    - a unique character string [input]
      * type          - MA type [input]
      * dim1/2        - array(dim1,dim2) as in FORTRAN [input]
      * chunk1/2      - minimum size that dimensions should
      *                 be chunked up into [input]
      *                 setting chunk1=dim1 gives distribution by rows
      *                 setting chunk2=dim2 gives distribution by columns 
      *                 Actual chunk sizes are modified so that they are
      *                 at least the min size and each process has either
      *                 zero or one chunk. 
      *                 chunk1/2 <=1 yields even distribution
      * g_a           - Integer handle for future references [output]
      */
{
int   i, nprocx, nprocy, fchunk1, fchunk2;
static Integer map1[MAX_NPROC], map2[MAX_NPROC];
Integer nblock1, nblock2;

      if(!GAinitialized) ga_error("GA not initialized ", 0);

      /* sync is in ga_create_irreg */

      if(*type != MT_F_DBL && *type != MT_F_INT &&  *type != MT_F_DCPL)
         ga_error("ga_create: type not yet supported ",  *type);
      else if( *dim1 <= 0 )
         ga_error("ga_create: array dimension1 invalid ",  *dim1);
      else if( *dim2 <= 0)
         ga_error("ga_create: array dimension2 invalid ",  *dim2);

      /*** figure out chunking ***/
      if(*chunk1 <= 1 && *chunk2 <= 1){
        if(*dim1 == 1)      { nprocx =1; nprocy=(int)GAnproc;}
        else if(*dim2 == 1) { nprocy =1; nprocx=(int)GAnproc;}
        else {
           /* nprocx= (int)sqrt((double)GAnproc);*/
           double dproc = ((double)GAnproc*(*dim1))/((double) *dim2);
           nprocx = NEAR_INT(sqrt(dproc)); 
           nprocx = MAX(1, nprocx); /* to avoid division by 0 */
           for(i=nprocx;i>0&& (GAnproc%i);i--);
           nprocx =(int)i; nprocy=(int)GAnproc/nprocx;
        }

        fchunk1 = (int) MAX(1, *dim1/nprocx);
        fchunk2 = (int) MAX(1, *dim2/nprocy);
      }else if(*chunk1 <= 1){
        fchunk1 = (int) MAX(1, (*dim1 * *dim2)/(GAnproc* *chunk2));
        fchunk2 = (int) *chunk2;
      }else if(*chunk2 <= 1){
        fchunk1 = (int) *chunk1;
        fchunk2 = (int) MAX(1, (*dim1 * *dim2)/(GAnproc* *chunk1));
      }else{
        fchunk1 = (int) MAX(1,  *chunk1);
        fchunk2 = (int) MAX(1,  *chunk2);
      }

      fchunk1 = (int)MIN(fchunk1, *dim1);
      fchunk2 = (int)MIN(fchunk2, *dim2);

      /*** chunk size correction for load balancing ***/
      while(((*dim1-1)/fchunk1+1)*((*dim2-1)/fchunk2+1) >GAnproc){
           if(fchunk1 == *dim1 && fchunk2 == *dim2) 
                     ga_error("ga_create: chunking failed !! ", 0L);
           if(fchunk1 < *dim1) fchunk1 ++; 
           if(fchunk2 < *dim2) fchunk2 ++; 
      }

      /* Now build map arrays */
      for(i=0, nblock1=0; i< *dim1; i += fchunk1, nblock1++) map1[nblock1]=i+1;
      for(i=0, nblock2=0; i< *dim2; i += fchunk2, nblock2++) map2[nblock2]=i+1;   
      if(GAme==0&& DEBUG){
         fprintf(stderr,"blocks (%d,%d)\n",nblock1, nblock2);
         fprintf(stderr,"chunks (%d,%d)\n",fchunk1, fchunk2);
         if(GAme==0){
           for (i=0;i<nblock1;i++)fprintf(stderr," %d ",map1[i]);
           for (i=0;i<nblock2;i++)fprintf(stderr," .%d ",map2[i]);
         }
      }

      return( ga_create_irreg(type, dim1, dim2, array_name, map1, &nblock1,
                         map2, &nblock2, g_a) );
}



/*\ CREATE A GLOBAL ARRAY
 *  Fortran version
\*/
#if defined(CRAY) || defined(WIN32)
logical FATR ga_create_(type, dim1, dim2, array_name, chunk1, chunk2, g_a)
     Integer *type, *dim1, *dim2, *chunk1, *chunk2, *g_a;
     _fcd array_name;
#else
logical ga_create_(type, dim1, dim2, array_name, chunk1, chunk2, g_a, slen)
     Integer *type, *dim1, *dim2, *chunk1, *chunk2, *g_a;
     char* array_name;
     int slen;
#endif
{
char buf[FNAM];
#if defined(CRAY) || defined(WIN32)
      f2cstring(_fcdtocp(array_name), _fcdlen(array_name), buf, FNAM);
#else
      f2cstring(array_name ,slen, buf, FNAM);
#endif

  return(ga_create(type, dim1, dim2, buf, chunk1, chunk2, g_a));
}



void gai_init_struct(handle)
Integer handle;
{
     if(!GA[handle].ptr){
        int len = MIN(GAnproc, MAX_PTR);
        GA[handle].ptr = (char**)malloc(len*sizeof(char**));
     }
     if(!GA[handle].mapc){
        int len = MAPLEN;
        GA[handle].mapc = (int*)malloc(len*sizeof(int*));
     }
     if(!GA[handle].ptr)ga_error("malloc failed: ptr:",0);
     if(!GA[handle].mapc)ga_error("malloc failed: mapc:",0);
}


/*\ get memory alligned w.r.t. MA base
 *  required on Linux as g77 ignores natural data alignment in common blocks
\*/ 
int gai_getmem(void **ptr_arr, int bytes, int type)
{
int status;
#ifndef CHECK_MA_ALGN
char *base;
long diff, item_size;  
Integer *adjust;
int i;

    /* need to enforce proper, natural allignment (on size boundary)  */
    switch (type){
      case MT_F_DBL: base =  (char *) DBL_MB; break;
      case MT_F_INT: base =  (char *) INT_MB; break;
      case MT_F_DCPL: base =  (char *) DCPL_MB; break;
      default:        base = (char*)0;
    }

    item_size = GAsizeofM(type);
    bytes += item_size; /***** will change on clusters *****/

#endif

    status = ARMCI_Malloc(ptr_arr, bytes);
    if(status) return status;

#ifndef CHECK_MA_ALGN

    /* adjust all addresses if they are not alligned on corresponding nodes*/
    adjust = (Integer*)gai_malloc(GAnproc*sizeof(Integer));

    diff = (ABS( base - (char *) ptr_arr[GAme])) % item_size; 
    for(i=0;i<GAnproc;i++)adjust[i]=0;
    adjust[GAme] = (diff > 0) ? item_size - diff : 0;

    ga_igop(GA_TYPE_GSM, adjust, GAnproc, "+");
    
    for(i=0;i<GAnproc;i++) ptr_arr[i] = adjust[i] + (char*)ptr_arr[i];

#endif
    return status;
}



/*\ CREATE A GLOBAL ARRAY -- IRREGULAR DISTRIBUTION
\*/
logical ga_create_irreg(type, dim1, dim2, array_name, map1, nblock1,
                         map2, nblock2, g_a)
     Integer *type, *dim1, *dim2, *map1, *map2, *nblock1, *nblock2, *g_a;
     char *array_name;
     /*
      * array_name    - a unique character string [input]
      * type          - MA type [input]
      * dim1/2        - array(dim1,dim2) as in FORTRAN [input]
      * nblock1       - no. of blocks first dimension is divided into [input]
      * nblock2       - no. of blocks second dimension is divided into [input]
      * map1          - no. ilo in each block [input]
      * map2          - no. jlo in each block [input]
      * g_a           - Integer handle for future references [output]
      */
{
char     op='*', *ptr = NULL;
Integer  ilo, ihi, jlo, jhi;
Integer  mem_size, nelem, mem_size_proc;
Integer  i, ga_handle, status;

      if(!GAinitialized) ga_error("GA not initialized ", 0);

      ga_sync_();

      GAstat.numcre ++; 

      if(*type != MT_F_DBL && *type != MT_F_INT &&  *type != MT_F_DCPL)
         ga_error("ga_create_irreg: type not yet supported ",  *type);
      else if( *dim1 <= 0 )
         ga_error("ga_create_irreg: array dimension1 invalid ",  *dim1);
      else if( *dim2 <= 0)
         ga_error("ga_create_irreg: array dimension2 invalid ",  *dim2);
      else if(*nblock1 <= 0)
         ga_error("ga_create_irreg: nblock1 <=0  ",  *nblock1);
      else if(*nblock2 <= 0)
         ga_error("ga_create_irreg: nblock2 <=0  ",  *nblock2);
      else if(*nblock1 * *nblock2 > GAnproc)
         ga_error("ga_create_irreg: too many blocks ",*nblock1 * *nblock2);

      if(GAme==0&& DEBUG){
        fprintf(stderr," array:%d map1:\n", *g_a);
        for (i=0;i<*nblock1;i++)fprintf(stderr," %d |",map1[i]);
        fprintf(stderr," \n array:%d map2:\n", *g_a);
        for (i=0;i<*nblock2;i++)fprintf(stderr," %d |",map2[i]);
        fprintf(stderr,"\n\n");
      }

      /*** Get next free global array handle ***/
      ga_handle =-1; i=0;
      do{
          if(!GA[i].actv) ga_handle=i;
          i++;
      }while(i<max_global_array && ga_handle==-1);
      if( ga_handle == -1)
          ga_error("ga_create: too many arrays ", (Integer)max_global_array);
      *g_a = (Integer)ga_handle - GA_OFFSET;

      /*** fill in Global Info Record for g_a ***/
      gai_init_struct(ga_handle);
      GA[ga_handle].type = *type;
      GA[ga_handle].actv = 1;
      strcpy(GA[ga_handle].name, array_name);
      GA[ga_handle].dims[0] = (int)*dim1;
      GA[ga_handle].dims[1] = (int)*dim2;
      GA[ga_handle].nblock[0] = (int) *nblock1;
      GA[ga_handle].nblock[1] = (int) *nblock2;
      GA[ga_handle].scale[0] = (double)*nblock1/(double)*dim1;
      GA[ga_handle].scale[1] = (double)*nblock2/(double)*dim2;

      /* Copy distribution maps, map1 & map2, into mapc:
       * . since nblock1*nblock2<=GAnproc,  mapc[GAnproc+1] suffices
       *   to pack everything into it;
       * . the dimension of block i is given as: MAX(mapc[i+1]-mapc[i],dim1/2)
       */
      for(i=0;i< *nblock1; i++) GA[ga_handle].mapc[i] = (int)map1[i];
      for(i=0;i< *nblock2; i++) GA[ga_handle].mapc[i+ *nblock1] = (int)map2[i];
      GA[ga_handle].mapc[*nblock1 + *nblock2] = -1; /* end of block marker */

      if(GAme ==0 && DEBUG){
         fprintf(stderr,"\nmapc %d elem\n", *nblock1 + *nblock2);
         for(i=0;i<1+*nblock1+ *nblock2;i++)
             fprintf(stderr,"%d,",GA[ga_handle].mapc[i]);
         fprintf(stderr,"\n\n");
      }


      /*** determine which portion of the array I am supposed to hold ***/
      ga_distribution_(g_a, &GAme, &ilo, &ihi, &jlo, &jhi);
      GA[ga_handle].chunk[0] = (int) (ihi-ilo+1);
      GA[ga_handle].chunk[1] = (int) (jhi-jlo+1);
      GA[ga_handle].ilo = ilo;
      GA[ga_handle].jlo = jlo;
      nelem = (ihi-ilo+1)*(jhi-jlo+1);

      mem_size = mem_size_proc =  nelem * GAsizeofM(*type);  
      GA[ga_handle].id = INVALID_MA_HANDLE;

      /* on shmem platforms, we use avg mem_size per processor in cluster */
      GA[ga_handle].size = mem_size_proc;

      /* if requested, enforce limits on memory consumption */
      if(GA_memory_limited) GA_total_memory -= mem_size_proc; 


      /* check if everybody has enough memory left */
      if(GA_memory_limited){
         status = (GA_total_memory >= 0) ? 1 : 0;
         ga_igop(GA_TYPE_GSM, &status, 1, "*");
      }status = 1;
 
      if(status){
          status = !gai_getmem(GA[ga_handle].ptr,mem_size,*type);
      }else{
          GA[ga_handle].ptr[GAme]=NULL;
      }

      ga_sync_();

#     ifdef GA_CREATE_INDEF
      if(status){
         Integer one = 1;
         if (GAme == 0) fprintf(stderr,"Initializing GA array%ld\n",*g_a);
         if(*type == MT_F_DBL){ 
             double bad = DBL_MAX;
             ga_fill_patch_(g_a, &one, dim1, &one, dim2, (Void *) &bad);
         } else if (*type == MT_F_INT) {
             Integer bad = (Integer) INT_MAX;
             ga_fill_patch_(g_a, &one, dim1, &one, dim2, (Void *) &bad);
         } else if (*type == MT_F_DCPL) {
             DoubleComplex bad = {DBL_MAX, DBL_MAX};
             ga_fill_patch_(g_a, &one, dim1, &one, dim2, (Void *) &bad);
         } else {
             ga_error("ga_create_irreg: type not yet supported ",  *type);
         }
      }
#     endif

      if(status){
         GAstat.curmem += GA[ga_handle].size;
         GAstat.maxmem  = MAX(GAstat.maxmem, GAstat.curmem);
         return(TRUE);
      }else{
         ga_destroy_(g_a);
         return(FALSE);
      }
}


/*\ CREATE A GLOBAL ARRAY -- IRREGULAR DISTRIBUTION
 *  Fortran version
\*/
#if defined(CRAY) || defined(WIN32)
logical ga_create_irreg_(type, dim1, dim2, array_name, map1, nblock1,
                         map2, nblock2, g_a)
     Integer *type, *dim1, *dim2, *map1, *map2, *nblock1, *nblock2, *g_a;
     _fcd array_name;
#else
logical ga_create_irreg_(type, dim1, dim2, array_name, map1, nblock1,
                         map2, nblock2, g_a, slen)
     Integer *type, *dim1, *dim2, *map1, *map2, *nblock1, *nblock2, *g_a;
     char *array_name;
     int slen;
#endif
{
char buf[FNAM];
#if defined(CRAY) || defined(WIN32)
      f2cstring(_fcdtocp(array_name), _fcdlen(array_name), buf, FNAM);
#else
      f2cstring(array_name ,slen, buf, FNAM);
#endif
  return( ga_create_irreg(type, dim1, dim2, buf, map1, nblock1,
                         map2, nblock2, g_a));
}



/*\ DUPLICATE A GLOBAL ARRAY
 *  -- new array g_b will have properties of g_a
\*/
logical ga_duplicate(g_a, g_b, array_name)
     Integer *g_a, *g_b;
     char *array_name;
     /*
      * array_name    - a character string [input]
      * g_a           - Integer handle for reference array [input]
      * g_b           - Integer handle for new array [output]
      */
{
char     op='*', *ptr = NULL, **save_ptr;
Integer  mem_size, mem_size_proc, nelem;
Integer  i, ga_handle, status;
int      *save_mapc;
#ifdef _CRAYMPP
char opadd='+';
Integer ncols,maxcols=0,*ptr_Integer,nlocks;
Integer ilo,ihi,jlo,jhi;
long **ptr_ptr_long,*ptr_long;
#endif


      ga_sync_();

      GAstat.numcre ++; 

      ga_check_handleM(g_a,"ga_duplicate");       

      /* find a free global_array handle for g_b */
      ga_handle =-1; i=0;
      do{
        if(!GA[i].actv) ga_handle=i;
        i++;
      }while(i<max_global_array && ga_handle==-1);
      if( ga_handle == -1)
          ga_error("ga_duplicate: too many arrays ", (Integer)max_global_array);
      *g_b = (Integer)ga_handle - GA_OFFSET;

      gai_init_struct(ga_handle);

      /*** copy content of the data structure ***/
      save_ptr = GA[ga_handle].ptr;
      save_mapc = GA[ga_handle].mapc;
      GA[ga_handle] = GA[GA_OFFSET + *g_a];
      strcpy(GA[ga_handle].name, array_name);
      GA[ga_handle].ptr = save_ptr;
      GA[ga_handle].mapc = save_mapc;
      for(i=0;i<MAPLEN; i++)GA[ga_handle].mapc[i] = GA[GA_OFFSET+ *g_a].mapc[i];

      /*** Memory Allocation & Initialization of GA Addressing Space ***/
      nelem = GA[ga_handle].chunk[1]*GA[ga_handle].chunk[0];
      mem_size = mem_size_proc = GA[ga_handle].size; 
      GA[ga_handle].id = INVALID_MA_HANDLE;

      /* if requested, enforce limits on memory consumption */
      if(GA_memory_limited) GA_total_memory -= mem_size_proc;

      if(!GA_memory_limited || GA_total_memory >= 0)
          status = !gai_getmem(GA[ga_handle].ptr,mem_size,GA[ga_handle].type);
      else{
          status = 0;
          GA[ga_handle].ptr[GAme]=NULL;
      }

      ga_sync_();

#     ifdef GA_CREATE_INDEF
      if(status){
         Integer one = 1; 
         Integer dim1 =GA[ga_handle].dims[1], dim2=GA[ga_handle].dims[2];
         if(GAme==0)fprintf(stderr,"duplicate:initializing GA array%ld\n",*g_b);
         if(GA[ga_handle].type == MT_F_DBL) {
             DoublePrecision bad = DBL_MAX;
             ga_fill_patch_(g_b, &one, &dim1, &one, &dim2,  &bad);
         } else if (GA[ga_handle].type == MT_F_INT) {
             Integer bad = (Integer) INT_MAX;
             ga_fill_patch_(g_b, &one, &dim1, &one, &dim2,  &bad);
         } else if (GA[ga_handle].type == MT_F_DCPL) {
             DoubleComplex bad = {DBL_MAX, DBL_MAX};
             ga_fill_patch_(g_b, &one, &dim1, &one, &dim2,  &bad);
         } else {
             ga_error("ga_duplicate: type not supported ",GA[ga_handle].type);
         }
      }
#     endif


      if(status){
         GAstat.curmem += GA[ga_handle].size;
         GAstat.maxmem  = MAX(GAstat.maxmem, GAstat.curmem);
         return(TRUE);
      }else{ 
         ga_destroy_(g_b);
         return(FALSE);
      }
}


/*\ DUPLICATE A GLOBAL ARRAY
 *  Fortran version
\*/
#if defined(CRAY) || defined(WIN32)
logical FATR ga_duplicate_(g_a, g_b, array_name)
     Integer *g_a, *g_b;
     _fcd array_name;
#else
logical FATR ga_duplicate_(g_a, g_b, array_name, slen)
     Integer *g_a, *g_b;
     char  *array_name;
     int   slen;
#endif
{
char buf[FNAM];
#if defined(CRAY) || defined(WIN32)
      f2cstring(_fcdtocp(array_name), _fcdlen(array_name), buf, FNAM);
#else
      f2cstring(array_name ,slen, buf, FNAM);
#endif

  return(ga_duplicate(g_a, g_b, buf));
}




/*\ DESTROY A GLOBAL ARRAY
\*/
logical FATR ga_destroy_(Integer *g_a)
{
Integer ga_handle = GA_OFFSET + *g_a;

    ga_sync_();
    GAstat.numdes ++; /*regardless of array status we count this call */

    /* fails if handle is out of range or array not active */
    if(ga_handle < 0 || ga_handle >= max_global_array) return FALSE;
    if(GA[ga_handle].actv==0) return FALSE;       
 
    ARMCI_Free(GA[ga_handle].ptr[GAme]);
    if(GA_memory_limited) GA_total_memory += GA[ga_handle].size;
    GA[ga_handle].actv = 0;     
    GAstat.curmem -= GA[ga_handle].size;

    return(TRUE);
}

    
     
/*\ TERMINATE GLOBAL ARRAY STRUCTURES
 *
 *  all GA arrays are destroyed & shared memory is dealocated
 *  GA routines (except for ga_initialize) should not be called thereafter 
\*/
void FATR  ga_terminate_() 
{
Integer i, handle;
extern double t_dgop, n_dgop, s_dgop;


    if(!GAinitialized) return;
    for (i=0;i<max_global_array;i++){
          handle = i - GA_OFFSET ;
          if(GA[i].actv) ga_destroy_(&handle);
          if(GA[i].ptr) free(GA[i].ptr);
          if(GA[i].mapc) free(GA[i].mapc);
    }
    ga_sync_();

    GA_total_memory = -1; /* restore "unlimited" memory usage status */
    GA_memory_limited = 0;

    ARMCI_Finalize();
    GAinitialized = 0;
}   

    
/*\ IS ARRAY ACTIVE/INACTIVE
\*/ 
Integer FATR ga_verify_handle_(g_a)
     Integer *g_a;
{
  return (Integer)
    ((*g_a + GA_OFFSET>= 0) && (*g_a + GA_OFFSET< max_global_array) && 
             GA[GA_OFFSET + (*g_a)].actv);
}
 


#define gaShmemLocation(proc, g_a, _i, _j, ptr_loc, ld)                        \
{                                                                              \
Integer _ilo, _ihi, _jlo, _jhi, offset, proc_place, g_handle=(g_a)+GA_OFFSET;  \
                                                                               \
      ga_ownsM(g_handle, (proc), _ilo, _ihi, _jlo, _jhi);                      \
      if((_i)<_ilo || (_i)>_ihi || (_j)<_jlo || (_j)>_jhi){                    \
          sprintf(err_string,"%s: p=%d invalid i/j (%d,%d) >< (%d:%d,%d:%d)",  \
                 "gaShmemLocation", proc, (_i),(_j), _ilo, _ihi, _jlo, _jhi);  \
          ga_error(err_string, g_a );                                          \
      }                                                                        \
      offset = ((_i) - _ilo) + (_ihi-_ilo+1)*((_j)-_jlo);                      \
                                                                               \
      /* find location of the proc in current cluster pointer array */         \
      proc_place =  proc;                                             \
      *(ptr_loc) = GA[g_handle].ptr[proc_place] +                              \
                   offset*GAsizeofM(GA[g_handle].type);                        \
      *(ld) = _ihi-_ilo+1;                                                     \
}


#define ga_check_regionM(g_a, ilo, ihi, jlo, jhi, string){                     \
   if (*(ilo) <= 0 || *(ihi) > GA[GA_OFFSET + *(g_a)].dims[0] ||               \
       *(jlo) <= 0 || *(jhi) > GA[GA_OFFSET + *(g_a)].dims[1] ||               \
       *(ihi) < *(ilo) ||  *(jhi) < *(jlo)){                                   \
       sprintf(err_string,"%s:request(%d:%d,%d:%d) out of range (1:%d,1:%d)",  \
               string, *(ilo), *(ihi), *(jlo), *(jhi),                         \
               GA[GA_OFFSET + *(g_a)].dims[0], GA[GA_OFFSET + *(g_a)].dims[1]);\
       ga_error(err_string, *(g_a));                                           \
   }                                                                           \
}


/*\ local put of a 2-dimensional patch of data into a global array
\*/
void ga_put_local(Integer g_a, Integer ilo, Integer ihi,
                               Integer jlo, Integer jhi,
                  void* buf, Integer offset, Integer ld_loc, Integer proc)
{
char     *ptr_glob, *ptr_loc;
int      ld_glob, rows, cols, size=GAsizeofM(GA[GA_OFFSET + g_a].type);
int      count[2];

   GA_PUSH_NAME("ga_put_local");

   rows = ihi - ilo +1;
   cols = jhi - jlo +1;

   gaShmemLocation(proc, g_a, ilo, jlo, &ptr_glob, &ld_glob);
   ptr_loc = (char *)buf  +  size* offset;
   count[0] = rows*size; count[1]=cols;
   ld_loc *= size; ld_glob *= size;

   ARMCI_PutS(ptr_loc, &ld_loc, ptr_glob, &ld_glob, count, 1, proc); 

   GA_POP_NAME;
}




/*\ PUT A 2-DIMENSIONAL PATCH OF DATA INTO A GLOBAL ARRAY 
\*/
void FATR  ga_put_(g_a, ilo, ihi, jlo, jhi, buf, ld)
   Integer  *g_a,  *ilo, *ihi, *jlo, *jhi,  *ld;
   Void  *buf;
{
Integer  p, np, proc, idx, type=GA[GA_OFFSET + *g_a].type;
Integer  ilop, ihip, jlop, jhip, offset, size, rows, cols;

#ifdef GA_TRACE
   trace_stime_();
#endif

      GA_PUSH_NAME("ga_put");

      size = GAsizeofM(type);
      GAstat.numput++;

      GAbytes.puttot += (double)size*(*ihi-*ilo+1)*(*jhi-*jlo+1);

      if(!ga_locate_region_(g_a, ilo, ihi, jlo, jhi, map, &np )){
          sprintf(err_string, "cannot locate region (%d:%d,%d:%d)",
                  *ilo, *ihi, *jlo, *jhi);
          ga_error(err_string, *g_a);
      }

      gaPermuteProcList(np); 
      for(idx=0; idx<np; idx++){
          p = (Integer)ProcListPerm[idx];

          ilop = map[p][0];
          ihip = map[p][1];
          jlop = map[p][2];
          jhip = map[p][3];
          proc = map[p][4];

          
          rows = ihip-ilop+1;
          cols = jhip-jlop+1;

          if(proc == GAme){
             GAbytes.putloc += (double)size*rows*cols;
          }

          offset = (jlop - *jlo)* *ld + ilop - *ilo;
          ga_put_local(*g_a, ilop, ihip, jlop, jhip, buf, offset, *ld, proc);

      }

      GA_POP_NAME;


#ifdef GA_TRACE
   trace_etime_();
   op_code = GA_OP_PUT; 
   trace_genrec_(g_a, ilo, ihi, jlo, jhi, &op_code);
#endif
}


/*\ local get of a 2-dimensional patch of data from a global array
\*/
void ga_get_local(Integer g_a, Integer ilo, Integer ihi, Integer jlo,
                  Integer jhi, void* buf, Integer offset, Integer ld_loc,
                  Integer proc)
{
char     *ptr_glob, *ptr_loc;
int      ld_glob, rows, cols, size=GAsizeofM(GA[GA_OFFSET + g_a].type);
int      count[2];

   GA_PUSH_NAME("ga_get_local");

   rows = ihi - ilo +1;
   cols = jhi - jlo +1;

   gaShmemLocation(proc, g_a, ilo, jlo, &ptr_glob, &ld_glob);

   ptr_loc = (char *)buf  + size * offset;
   count[0] = rows*size; count[1]=cols;
   ld_loc *= size; ld_glob *= size;

   ARMCI_GetS(ptr_glob, &ld_glob, ptr_loc, &ld_loc, count, 1, proc); 

   GA_POP_NAME;
}



/*\ GET A 2-DIMENSIONAL PATCH OF DATA FROM A GLOBAL ARRAY
\*/
void FATR  ga_get_(g_a, ilo, ihi, jlo, jhi, buf, ld)
   Integer  *g_a, *ilo, *ihi, *jlo, *jhi,  *ld;
   Void     *buf;
{
Integer p, np, proc, idx, type=GA[GA_OFFSET + *g_a].type;
Integer ilop, ihip, jlop, jhip, offset, size, rows,cols;

#ifdef GA_TRACE
   trace_stime_();
#endif

      GA_PUSH_NAME("ga_get");
      GAstat.numget++;
      size = GAsizeofM(type);
      GAbytes.gettot += (double)size*(*ihi-*ilo+1)*(*jhi-*jlo+1);

      if(!ga_locate_region_(g_a, ilo, ihi, jlo, jhi, map, &np )){
          sprintf(err_string, "cannot locate region (%d:%d,%d:%d)",
                  *ilo, *ihi, *jlo, *jhi);
          ga_error(err_string, *g_a);
      }


      gaPermuteProcList(np);

      for(idx=0; idx<np; idx++){
          p = (Integer)ProcListPerm[idx];
          ilop = map[p][0];
          ihip = map[p][1];
          jlop = map[p][2];
          jhip = map[p][3];
          proc = map[p][4]; 


          rows = ihip-ilop+1;
          cols = jhip-jlop+1;

          if(proc == GAme){
             GAbytes.getloc += (double)size*rows*cols;
          }

          offset = (jlop - *jlo)* *ld + ilop - *ilo;
          ga_get_local(*g_a, ilop, ihip, jlop, jhip, buf, offset, *ld, proc);

      }

      GA_POP_NAME;

#ifdef GA_TRACE
   trace_etime_();
   op_code = GA_OP_GET; 
   trace_genrec_(g_a, ilo, ihi, jlo, jhi, &op_code);
#endif
}



/*\ local accumulate
\*/
void ga_acc_local(Integer g_a, Integer ilo, Integer ihi, Integer jlo, 
                  Integer jhi, void* buf, Integer offset, Integer ld_loc,
                  Integer proc, void* alpha)
{
char     *ptr_loc, *ptr_glob;
int      ld_glob, rows, cols, type = GA[GA_OFFSET + g_a].type;
int      size=GAsizeofM(type);
int      count[2], optype;

   GA_PUSH_NAME("ga_acc_local"); 

   gaShmemLocation(proc, g_a, ilo, jlo, &ptr_glob, &ld_glob);
   ptr_loc = (char *)buf   + size * offset;
   rows = ihi - ilo +1;
   cols = jhi - jlo +1;

   count[0] = rows*size; count[1]=cols;
   ld_loc *= size; ld_glob *= size;

   if(GA[GA_OFFSET + g_a].type==MT_F_DBL) optype= ARMCI_ACC_DBL;
   else if(type==MT_F_DCPL)optype= ARMCI_ACC_DCP;
   else if(size==sizeof(int))optype= ARMCI_ACC_INT;
   else ga_error("now is the time to add MT_F_LONG to armci!",0);

   ARMCI_AccS(optype, alpha, ptr_loc, &ld_loc, ptr_glob, &ld_glob, count, 1,proc);

   GA_POP_NAME;

}



/*\ ACCUMULATE OPERATION FOR A 2-DIMENSIONAL PATCH OF GLOBAL ARRAY
 *
 *  g_a += alpha * patch
\*/
void FATR  ga_acc_(g_a, ilo, ihi, jlo, jhi, buf, ld, alpha)
   Integer *g_a, *ilo, *ihi, *jlo, *jhi, *ld;
   void *buf, *alpha;
{
   Integer ilop, ihip, jlop, jhip, offset, type = GA[GA_OFFSET + *g_a].type;
   Integer p, np, proc, idx, rows, cols, size = GAsizeofM(type);

#ifdef GA_TRACE
   trace_stime_();
#endif

   GA_PUSH_NAME("ga_acc");
   GAstat.numacc++;
   GAbytes.acctot += (double)size*(*ihi-*ilo+1)*(*jhi-*jlo+1);


   if(!ga_locate_region_(g_a, ilo, ihi, jlo, jhi, map, &np )){
          sprintf(err_string, "cannot locate region (%d:%d,%d:%d)",
                  *ilo, *ihi, *jlo, *jhi);
          ga_error(err_string, *g_a);
   }

   gaPermuteProcList(np); /* prepare permuted list of indices */

   for(idx=0; idx<np; idx++){
       p = (Integer)ProcListPerm[idx];

       ilop = map[p][0];
       ihip = map[p][1];
       jlop = map[p][2];
       jhip = map[p][3];
       proc = map[p][4];

       if(proc == GAme){

             rows = ihip-ilop+1;
             cols = jhip-jlop+1;
             GAbytes.accloc += (double)size*rows*cols;

        }


        offset = (jlop - *jlo)* *ld + ilop - *ilo;
        ga_acc_local(*g_a, ilop, ihip, jlop, jhip, buf, offset, *ld,
                       proc, alpha);

   }

  GA_POP_NAME;

#ifdef GA_TRACE
   trace_etime_();
   op_code = GA_OP_ACC; 
   trace_genrec_(g_a, ilo, ihi, jlo, jhi, &op_code);
#endif
}



/*\ PROVIDE ACCESS TO A PATCH OF A GLOBAL ARRAY
\*/
void FATR  ga_access_(g_a, ilo, ihi, jlo, jhi, index, ld)
   Integer *g_a, *ilo, *ihi, *jlo, *jhi, *index, *ld;
{
register char *ptr;
Integer  item_size, proc_place, handle = GA_OFFSET + *g_a;

Integer ow;

   if(!ga_locate_(g_a,ilo, jlo, &ow))ga_error("ga_access:locate top failed",0);
   if(ow != GAme) ga_error("ga_access: cannot access top of the patch",0);
   if(!ga_locate_(g_a,ihi, jhi, &ow))ga_error("ga_access:locate bot failed",0);
   if(ow != GAme) ga_error("ga_access: cannot access bottom of the patch",0);

   ga_check_handleM(g_a, "ga_access");
   ga_check_regionM(g_a, ilo, ihi, jlo, jhi, "ga_access");

   item_size = (int) GAsizeofM(GA[GA_OFFSET + *g_a].type);

   proc_place = GAme -  GAmaster;

   ptr = GA[handle].ptr[proc_place] + item_size * ( (*jlo - GA[handle].jlo )
         *GA[handle].chunk[0] + *ilo - GA[handle].ilo);
   *ld    = GA[handle].chunk[0];  
   FLUSH_CACHE;


   /*
    * return address of the patch  as the distance in bytes
    * from the reference address
    *
    * .in Fortran we need only the index to the type array: dbl_mb or int_mb
    *  that are elements of COMMON in the the mafdecls.h include file
    *
    * .in C we need both the index and the pointer
    *
    */ 
   /* compute index and check if it is correct */
   switch (GA[handle].type){
     case MT_F_DBL: 
        *index = (Integer) (ptr - (char*)DBL_MB);
        if(ptr != ((char*)DBL_MB)+ *index ){ 
               ga_error("ga_access: MA addressing problem dbl - index",handle);
        }
        break;

     case MT_F_DCPL:
        *index = (Integer) (ptr - (char*)DCPL_MB);
        if(ptr != ((char*)DCPL_MB)+ *index ){
              ga_error("ga_access: MA addressing problem dcpl - index",handle);
        }
        break;

     case MT_F_INT:
        *index = (Integer) (ptr - (char*)INT_MB);
        if(ptr != ((char*)INT_MB) + *index) {
               ga_error("ga_access: MA addressing problem int - index",handle);
        }
        break;

   }

   /* check the allignment */
   if(*index % item_size){
       fprintf(stderr,"index=%ld size=%ld off =%ld\n",(long)*index, (long)item_size,(long)*index%item_size);
       ga_error(" ga_access: base address misallignment ",(long)index);
   }

   /* adjust index according to the data type */
   *index /= item_size;

   /* adjust index for Fortran addressing */
   (*index) ++ ;
}



/*\ RELEASE ACCESS TO A PATCH OF A GLOBAL ARRAY
\*/
void FATR  ga_release_(g_a, ilo, ihi, jlo, jhi)
     Integer *g_a, *ilo, *ihi, *jlo, *jhi;
{}


/*\ RELEASE ACCESS & UPDATE A PATCH OF A GLOBAL ARRAY
\*/
void FATR  ga_release_update_(g_a, ilo, ihi, jlo, jhi)
     Integer *g_a, *ilo, *ihi, *jlo, *jhi;
{}



/*\ INQUIRE POPERTIES OF A GLOBAL ARRAY
 *  Fortran version
\*/ 
void FATR  ga_inquire_(g_a,  type, dim1, dim2)
      Integer *g_a, *dim1, *dim2, *type;
{
   ga_check_handleM(g_a, "ga_inquire");
   *type       = GA[GA_OFFSET + *g_a].type;
   *dim1       = GA[GA_OFFSET + *g_a].dims[0];
   *dim2       = GA[GA_OFFSET + *g_a].dims[1];
}



/*\ INQUIRE NAME OF A GLOBAL ARRAY
 *  Fortran version
\*/
#if defined(CRAY) || defined(WIN32)
void FATR  ga_inquire_name_(g_a, array_name)
      Integer *g_a;
      _fcd    array_name;
{
   c2fstring(GA[GA_OFFSET+ *g_a].name,_fcdtocp(array_name),_fcdlen(array_name));
}
#else
void FATR  ga_inquire_name_(g_a, array_name, len)
      Integer *g_a;
      char    *array_name;
      int     len;
{
   c2fstring(GA[GA_OFFSET + *g_a].name, array_name, len);
}
#endif




/*\ INQUIRE NAME OF A GLOBAL ARRAY
 *  C version
\*/
void ga_inquire_name(g_a, array_name)
      Integer *g_a;
      char    *array_name;
{ 
   ga_check_handleM(g_a, "ga_inquire_name");
   strcpy(array_name, GA[GA_OFFSET + *g_a].name);
}



/*\ RETURN COORDINATES OF A GA PATCH ASSOCIATED WITH CALLING PROCESSOR
\*/
void FATR  ga_distribution_(g_a, proc, ilo, ihi, jlo, jhi)
   Integer *g_a, *ilo, *ihi, *jlo, *jhi, *proc;
{
register Integer iproc, jproc, loc, ga_handle;

   ga_check_handleM(g_a, "ga_distribution");

   ga_handle = (GA_OFFSET + *g_a);
 
   if(*proc > GA[ga_handle].nblock[0] * GA[ga_handle].nblock[1] - 1 || *proc<0){
         *ilo = (Integer)0;    *jlo = (Integer)0; 
         *ihi = (Integer)-1;   *jhi = (Integer)-1; 
   }else{
         jproc = (*proc)/GA[ga_handle].nblock[0]; 
         iproc = (*proc)%GA[ga_handle].nblock[0]; 
         loc = iproc;

         *ilo = GA[ga_handle].mapc[loc]; *ihi = GA[ga_handle].mapc[loc+1] -1; 

         /* correction to find the right spot in mapc*/
         loc = jproc + GA[ga_handle].nblock[0];
         *jlo = GA[ga_handle].mapc[loc]; *jhi = GA[ga_handle].mapc[loc+1] -1; 

         if( iproc == GA[ga_handle].nblock[0] -1) *ihi = GA[ga_handle].dims[0];
         if( jproc == GA[ga_handle].nblock[1] -1) *jhi = GA[ga_handle].dims[1];
   }
}



/*\ RETURN COORDINATES OF ARRAY BLOCK HELD BY A PROCESSOR
\*/
void FATR  ga_proc_topology_(g_a, proc, pr, pc)
   Integer *g_a, *proc, *pr, *pc;
{
register Integer ga_handle;

   ga_check_handleM(g_a, "ga_proc_topology");

   ga_handle = (GA_OFFSET + *g_a);

   if(*proc > GA[ga_handle].nblock[0] * GA[ga_handle].nblock[1] - 1 || *proc<0){
         *pc = -1; *pr = -1;
   }else{
         *pc = (*proc)/GA[ga_handle].nblock[0];
         *pr = (*proc)%GA[ga_handle].nblock[0];
   }
}



/*\ finds block i that 'elem' belongs to: map[i]>= elem < map[i+1]
\*/
int findblock2(map,n,scale,elem)
    int *map, n, elem;
    double scale;
{
int candidate, found, b;

    candidate = (int)(scale*elem)-1;
    if(candidate<0)candidate =0;
    found = 0;
    if(map[candidate]>elem){ /* search downward */
         b= candidate-1;
         while(b>=0){
            found = (map[b]<=elem);
            if(found)break;
            b--;
         } 
    }else{ /* search upward */
         b= candidate;
         while(b<n-1){ 
            found = (map[b+1]>elem);
            if(found)break;
            b++;
         }
    }
    if(!found)b=n-1;
    return(b);
}

        
#define findblock(map_ij,n,scale,elem, block)\
{\
int candidate, found, b, *map= (map_ij);\
\
    candidate = (int)(scale*(elem));\
    found = 0;\
    if(map[candidate] <= (elem)){ /* search downward */\
         b= candidate;\
         while(b<(n)-1){ \
            found = (map[b+1]>(elem));\
            if(found)break;\
            b++;\
         } \
    }else{ /* search upward */\
         b= candidate-1;\
         while(b>=0){\
            found = (map[b]<=(elem));\
            if(found)break;\
            b--;\
         }\
    }\
    if(!found)b=(n)-1;\
    *(block) = b;\
}



/*\ LOCATE THE OWNER OF THE (i,j) ELEMENT OF A GLOBAL ARRAY
\*/
logical FATR ga_locate_(g_a, i, j, owner)
        Integer *g_a, *i, *j, *owner;
{
int  iproc, jproc; 
Integer ga_handle;

   ga_check_handleM(g_a, "ga_locate");
   ga_handle = GA_OFFSET + *g_a;

   if (*i <= 0 || *i > GA[GA_OFFSET + *g_a].dims[0]  ||
       *j <= 0 || *j > GA[GA_OFFSET + *g_a].dims[1]){
       *owner = -1;
       return( FALSE);
   }

   findblock(GA[ga_handle].mapc,GA[ga_handle].nblock[0], 
             GA[ga_handle].scale[0], (int)*i, &iproc);
   findblock(GA[ga_handle].mapc+GA[ga_handle].nblock[0],
             GA[ga_handle].nblock[1], GA[ga_handle].scale[1], (int)*j,&jproc);

   *owner = (Integer)jproc* GA[ga_handle].nblock[0] + iproc;
   return(TRUE);
}



/*\ LOCATE OWNERS OF THE SPECIFIED PATCH OF A GLOBAL ARRAY
\*/
logical FATR ga_locate_region_(g_a, ilo, ihi, jlo, jhi, map, np )
        Integer *g_a, *ilo, *jlo, *ihi, *jhi, map[][5], *np;
{
int  iprocLT, iprocRB, jprocLT, jprocRB;
Integer  owner, ilop, ihip, jlop, jhip, i,j, ga_handle;

   ga_check_handleM(g_a, "ga_locate_region");

   ga_handle = GA_OFFSET + *g_a;
   if (*ilo <= 0 || *ihi > GA[ga_handle].dims[0] ||
       *jlo <= 0 || *jhi > GA[ga_handle].dims[1] ||
       *ihi < *ilo ||  *jhi < *jlo){
       return(FALSE);
   }

   /* find "processor coordinates" for the left top corner */
   findblock(GA[ga_handle].mapc,GA[ga_handle].nblock[0], 
             GA[ga_handle].scale[0],(int)*ilo, &iprocLT);
   findblock(GA[ga_handle].mapc+GA[ga_handle].nblock[0],
             GA[ga_handle].nblock[1],GA[ga_handle].scale[1],(int)*jlo,&jprocLT);

   /* find "processor coordinates" for the right bottom corner */
   findblock(GA[ga_handle].mapc,GA[ga_handle].nblock[0], 
             GA[ga_handle].scale[0],(int)*ihi, &iprocRB);
   findblock(GA[ga_handle].mapc+GA[ga_handle].nblock[0],
             GA[ga_handle].nblock[1],GA[ga_handle].scale[1],(int)*jhi,&jprocRB);

   *np = 0;
   for(i=iprocLT;i<=iprocRB;i++)
       for(j=jprocLT;j<=jprocRB;j++){
           owner = j* GA[ga_handle].nblock[0] + i;
           ga_ownsM(ga_handle, owner, ilop, ihip, jlop, jhip);
           map[*np][0] = *ilo < ilop ? ilop : *ilo;    
           map[*np][1] = *ihi > ihip ? ihip : *ihi;    
           map[*np][2] = *jlo < jlop ? jlop : *jlo;    
           map[*np][3] = *jhi > jhip ? jhip : *jhi;    
           map[*np][4] = owner;    
           (*np)++;
   }

   return(TRUE);
}
    

void ga_scatter_local(Integer g_a, Void *v,Integer *i,Integer *j,
                      Integer nv, Integer proc) 
{
     ga_error("ga_scatter_local should not be used",0);
}



void ga_scatter_acc_local(Integer g_a, Void *v,Integer *i,Integer *j,
                          Integer nv, void* alpha, Integer proc) 
{
void **ptr_src, **ptr_dst;
char *ptr_ref;
Integer ldp, item_size, ilo, ihi, jlo, jhi, type;
armci_giov_t desc;
register Integer k, offset;
int rc;

  if (nv < 1) return;

  GA_PUSH_NAME("ga_scatter_local");

  ga_distribution_(&g_a, &proc, &ilo, &ihi, &jlo, &jhi);

  /* get address of the first element owned by proc */
  gaShmemLocation(proc, g_a, ilo, jlo, &ptr_ref, &ldp);

  type = GA[GA_OFFSET + g_a].type;
  item_size = GAsizeofM(type);

  ptr_src = gai_malloc(2*nv*sizeof(void*));
  if(ptr_src==NULL)ga_error("gai_malloc failed",nv);
  else ptr_dst=ptr_src+ nv;

  for(k=0; k< nv; k++){
     if(i[k] < ilo || i[k] > ihi  || j[k] < jlo || j[k] > jhi){
       sprintf(err_string,"proc=%d invalid i/j=(%d,%d)>< [%d:%d,%d:%d]",
               proc, i[k], j[k], ilo, ihi, jlo, jhi); 
       ga_error(err_string,g_a);
     }

     offset  = (j[k] - jlo)* ldp + i[k] - ilo;
     ptr_dst[k] = ptr_ref + item_size * offset;
     ptr_src[k] = ((char*)v) + k*item_size;
  }
  desc.bytes = item_size;
  desc.src_ptr_array = ptr_src;
  desc.dst_ptr_array = ptr_dst;
  desc.ptr_array_len = nv;

  if(alpha ==NULL) rc=ARMCI_PutV(&desc, 1, (int)proc);
  else{

    int optype;
    if(type==MT_F_DBL) optype= ARMCI_ACC_DBL;
    else if(type==MT_F_DCPL)optype= ARMCI_ACC_DCP;
    else if(item_size==sizeof(int))optype= ARMCI_ACC_INT;
    else ga_error("now is the time to add MT_F_LONG to ARMCI!",0);
    rc= ARMCI_AccV(optype, alpha, &desc, 1, (int)proc);
  }

  if(rc) ga_error("scatter/_acc failed in armci",rc);

  gai_free(ptr_src);

  GA_POP_NAME;
}




/*\ SCATTER OPERATION elements of v into the global array
\*/
void FATR  ga_scatter_(g_a, v, i, j, nv)
     Integer *g_a, *nv, *i, *j;
     Void *v;
{
register Integer k;
Integer pindex, phandle, item_size;
Integer first, nelem, proc, type=GA[GA_OFFSET + *g_a].type;

  if (*nv < 1) return;

  ga_check_handleM(g_a, "ga_scatter");
  GA_PUSH_NAME("ga_scatter");
  GAstat.numsca++;

  if(!MA_push_get(MT_F_INT,*nv, "ga_scatter--p", &phandle, &pindex))
            ga_error("MA alloc failed ", *g_a);

  /* find proc that owns the (i,j) element; store it in temp: INT_MB[] */
  for(k=0; k< *nv; k++) if(! ga_locate_(g_a, i+k, j+k, INT_MB+pindex+k)){
         sprintf(err_string,"invalid i/j=(%d,%d)", i[k], j[k]);
         ga_error(err_string,*g_a);
  }

  /* determine limit for message size --  v,i, & j will travel together */
  item_size = GAsizeofM(type);
  GAbytes.scatot += (double)item_size**nv ;

  /* Sort the entries by processor */
  ga_sort_scat(nv, (DoublePrecision*)v, i, j, INT_MB+pindex, type );

  /* go through the list again executing scatter for each processor */

  first = 0;
  do { 
      proc  = INT_MB[pindex+first]; 
      nelem = 0;

      /* count entries for proc from "first" to last */
      for(k=first; k< *nv; k++){
        if(proc == INT_MB[pindex+k]) nelem++;
        else break;
      }

      if(proc == GAme){
             GAbytes.scaloc += (double)item_size* nelem ;
      }

      ga_scatter_acc_local(*g_a, ((char*)v)+item_size*first, i+first, 
                        j+first, nelem, NULL, proc);
      first += nelem;

  }while (first< *nv);

  if(! MA_pop_stack(phandle)) ga_error(" pop stack failed!",phandle);

  GA_POP_NAME;
}
      


/*\ SCATTER OPERATION elements of v into the global array
\*/
void FATR  ga_scatter_acc_(g_a, v, i, j, nv, alpha)
     Integer *g_a, *nv, *i, *j;
     Void *v, *alpha;
{
register Integer k;
Integer pindex, phandle, item_size;
Integer first, nelem, proc, type=GA[GA_OFFSET + *g_a].type;

  if (*nv < 1) return;

  ga_check_handleM(g_a, "ga_scatter_acc");
  GA_PUSH_NAME("ga_scatter_acc");
  GAstat.numsca++;

  if(!MA_push_get(MT_F_INT,*nv, "ga_scatter_acc--p", &phandle, &pindex))
            ga_error("MA alloc failed ", *g_a);

  /* find proc that owns the (i,j) element; store it in temp: INT_MB[] */
  for(k=0; k< *nv; k++) if(! ga_locate_(g_a, i+k, j+k, INT_MB+pindex+k)){
         sprintf(err_string,"invalid i/j=(%d,%d)", i[k], j[k]);
         ga_error(err_string,*g_a);
  }

  /* determine limit for message size --  v,i, & j will travel together */
  item_size = GAsizeofM(type);
  GAbytes.scatot += (double)item_size**nv ;

  /* Sort the entries by processor */
  ga_sort_scat(nv, (DoublePrecision*)v, i, j, INT_MB+pindex, type );

  /* go through the list again executing scatter for each processor */

  first = 0;
  do {
      proc  = INT_MB[pindex+first];
      nelem = 0;

      /* count entries for proc from "first" to last */
      for(k=first; k< *nv; k++){
        if(proc == INT_MB[pindex+k]) nelem++;
        else break;
      }

      if(proc == GAme){
             GAbytes.scaloc += (double)item_size* nelem ;
      }

      ga_scatter_acc_local(*g_a, ((char*)v)+item_size*first, i+first,
                           j+first, nelem, alpha, proc);
      first += nelem;

  }while (first< *nv);

  if(! MA_pop_stack(phandle)) ga_error(" pop stack failed!",phandle);

  GA_POP_NAME;
}



/*\ permutes input index list using sort routine used in scatter/gather
\*/
void FATR  ga_sort_permut_(g_a, index, i, j, nv)
     Integer *g_a, *nv, *i, *j, *index;
{
register Integer k;
Integer pindex, phandle;
extern void ga_sort_permutation();

  if (*nv < 1) return;

  if(!MA_push_get(MT_F_INT,*nv, "ga_sort_permut--p", &phandle, &pindex))
            ga_error("MA alloc failed ", *g_a);

  /* find proc that owns the (i,j) element; store it in temp: INT_MB[] */
  for(k=0; k< *nv; k++) if(! ga_locate_(g_a, i+k, j+k, INT_MB+pindex+k)){
         sprintf(err_string,"invalid i/j=(%d,%d)", i[k], j[k]);
         ga_error(err_string,*g_a);
  }

  /* Sort the entries by processor */
  ga_sort_permutation(nv, index, INT_MB+pindex);
  if(! MA_pop_stack(phandle)) ga_error(" pop stack failed!",phandle);
}



void ga_gather_local(g_a, v, i, j, nv, proc) 
     Integer g_a, *i, *j, nv, proc;
     Void *v;
{
void **ptr_src, **ptr_dst;
char *ptr_ref;
Integer ldp, item_size, ilo, ihi, jlo, jhi;
armci_giov_t desc;
register Integer k, offset;

  if (nv < 1) return;

  GA_PUSH_NAME("ga_gather_local");

  ga_distribution_(&g_a, &proc, &ilo, &ihi, &jlo, &jhi);

  /* get address of the first element owned by proc */
  gaShmemLocation(proc, g_a, ilo, jlo, &ptr_ref, &ldp);

  item_size = GAsizeofM(GA[GA_OFFSET + g_a].type);
  ptr_src = gai_malloc(2*nv*sizeof(void*));
  if(ptr_src==NULL)ga_error("gai_malloc failed",nv);
  else ptr_dst=ptr_src+ nv;

  for(k=0; k< nv; k++){
     if(i[k] < ilo || i[k] > ihi  || j[k] < jlo || j[k] > jhi){
       sprintf(err_string,"proc=%d invalid i/j=(%d,%d)>< [%d:%d,%d:%d]",
               proc, i[k], j[k], ilo, ihi, jlo, jhi); 
       ga_error(err_string,g_a);
     }

     offset  = (j[k] - jlo)* ldp + i[k] - ilo;
     ptr_dst[k] = ((char*)v) + k*item_size;
     ptr_src[k] = ptr_ref + item_size * offset;
  }
  desc.bytes = item_size;
  desc.src_ptr_array = ptr_src;
  desc.dst_ptr_array = ptr_dst;
  desc.ptr_array_len = nv;

  ARMCI_GetV(&desc, 1, (int)proc);

  gai_free(ptr_src);

  GA_POP_NAME;
}



/*\ GATHER OPERATION elements from the global array into v
\*/
void FATR  ga_gather_(g_a, v, i, j, nv)
     Integer *g_a, *nv, *i, *j;
     Void *v;
{
register Integer k, nelem;
Integer pindex, phandle, item_size;
Integer first, proc;

  if (*nv < 1) return;

  ga_check_handleM(g_a, "ga_gather");
  GA_PUSH_NAME("ga_gather");
  GAstat.numgat++;


  if(!MA_push_get(MT_F_INT, *nv, "ga_gather--p", &phandle, &pindex))
            ga_error("MA failed ", *g_a);

  /* find proc that owns the (i,j) element; store it in temp: INT_MB[] */
  for(k=0; k< *nv; k++) if(! ga_locate_(g_a, i+k, j+k, INT_MB+pindex+k)){
       sprintf(err_string,"invalid i/j=(%d,%d)", i[k], j[k]);
       ga_error(err_string,*g_a);
  }

  /* Sort the entries by processor */
  ga_sort_gath_(nv, i, j, INT_MB+pindex);
   
  item_size = GAsizeofM(GA[GA_OFFSET + *g_a].type);
  GAbytes.gattot += (double)item_size**nv;


  /* go through the list again executing gather for each processor */

  first = 0;
  do {
      proc  = INT_MB[pindex+first];
      nelem = 0;

      /* count entries for proc from "first" to last */
      for(k=first; k< *nv; k++){
        if(proc == INT_MB[pindex+k]) nelem++;
        else break;
      }

      if(proc == GAme){
             GAbytes.gatloc += (double)item_size* nelem ;
      }

      ga_gather_local(*g_a, ((char*)v)+item_size*first, i+first, j+first,
                        nelem,proc);
      first += nelem; 

  }while (first< *nv);

  if(! MA_pop_stack(phandle)) ga_error(" pop stack failed!",phandle);

  GA_POP_NAME;
}
      
           

/*\ local read and increment of an element of a global array
\*/
Integer ga_read_inc_local(g_a, i, j, inc, proc)
        Integer g_a, i, j, inc, proc;
{
Integer *ptr, ldp, value;
int optype;

   GA_PUSH_NAME("ga_read_inc_local");

   /* get a address of the g_a(i,j) element */
   gaShmemLocation(proc, g_a, i, j, (char**)&ptr, &ldp);

#  ifdef EXT_INT
     optype = ARMCI_FETCH_AND_ADD_LONG;
#  else
     optype = ARMCI_FETCH_AND_ADD;
#  endif

   ARMCI_Rmw(optype, (int*)&value, (int*)ptr, (int)inc, proc);
   
   GA_POP_NAME;

   return(value);
}



/*\ READ AND INCREMENT AN ELEMENT OF A GLOBAL ARRAY
\*/
Integer FATR ga_read_inc_(g_a, i, j, inc)
        Integer *g_a, *i, *j, *inc;
{
Integer  value, proc; 
#ifdef GA_TRACE
       trace_stime_();
#endif

    ga_check_handleM(g_a, "ga_read_inc");
    GA_PUSH_NAME("ga_read_inc");

    GAstat.numrdi++;
    GAbytes.rditot += (double)sizeof(Integer);

    if(GA[GA_OFFSET + *g_a].type !=MT_F_INT)
       ga_error("type must be integer ",*g_a);

    ga_locate_(g_a, i, j, &proc);
    value = ga_read_inc_local(*g_a, *i, *j, *inc, proc);

    if(GAme == proc)GAbytes.rdiloc += (double)sizeof(Integer);

#  ifdef GA_TRACE
     trace_etime_();
     op_code = GA_OP_RDI; 
     trace_genrec_(g_a, i, i, j, j, &op_code);
#  endif

   GA_POP_NAME;
   return(value);
}


Integer FATR ga_nodeid_()
{
  return ((Integer)GAme);
}


Integer FATR ga_nnodes_()
{
  return ((Integer)GAnproc);
}


logical FATR ga_compare_distr_(g_a, g_b)
     Integer *g_a, *g_b;
{
int h_a =*g_a + GA_OFFSET;
int h_b =*g_b + GA_OFFSET;
int i;

   GA_PUSH_NAME("ga_compare_distr");
   ga_check_handleM(g_a, "distribution a");
   ga_check_handleM(g_b, "distribution b");

   GA_POP_NAME;

   if(GA[h_a].dims[0] == GA[h_b].dims[0]) return FALSE;
   if(GA[h_a].dims[1] == GA[h_b].dims[1]) return FALSE;
   for(i=0; i <MAPLEN; i++){
      if(GA[h_a].mapc[i] != GA[h_b].mapc[i]) return FALSE;
      if(GA[h_a].mapc[i] == -1) break;
   }
   return TRUE;
}
/*************************************************************************/
